#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - DFP Configuration

These configuration utilities outline the way configurations may be
loaded into Parselmouth for authenticating different advertising client
APIs
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Third Party Imports
import yaml

# Parselmouth Imports
from parselmouth.exceptions import ParselmouthException
from parselmouth.adapters.abstract_config import AbstractConfig


class DFPConfig(AbstractConfig):
    """
    Configuration container specific to the DFP API
    """

    _credentials = {
        'client_id': None,
        'client_secret': None,
        'refresh_token': None,
        'application_name': None,
        'network_code': None,
    }
    """
    dict, key/value pairs required to connect to DFP service
    """

    def validate_credentials(self):
        """
        Make sure all credentials have been populated correctly
        """
        for _key, _val in self._credentials.items():
            if _val is None:
                raise ParselmouthException("{} field is missing".format(_key))

    def get_credentials_arguments(self):
        """
        @return: dict
        """
        return self._credentials

    def load_config_from_file(self, provider_name, config_path):

        with open(config_path, 'r') as infile:
            config_dict = yaml.load(infile)

        self._credentials = {
            'client_id': config_dict[provider_name]['client_id'],
            'client_secret': config_dict[provider_name]['client_secret'],
            'refresh_token': config_dict[provider_name]['refresh_token'],
            'network_code': config_dict[provider_name]['network_code'],
            'application_name': config_dict[provider_name]['application_name'],
        }

    def load_config_from_external_source(self,
                                         client_id,
                                         client_secret,
                                         refresh_token,
                                         application_name,
                                         network_code):
        """
        https://developers.google.com/doubleclick-publishers/docs/authentication

        @param client_id: str
        @param client_secret: str
        @param refresh_token: str
        @param application_name: str
        @param network_code: str
        @param version: str
        """
        self._credentials = {
            'client_id': client_id,
            'client_secret': client_secret,
            'refresh_token': refresh_token,
            'application_name': application_name,
            'network_code': network_code,
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - DFP Targeting Utilities
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
from collections import defaultdict

# Parselmouth Imports
from parselmouth.constants import TechnologyTargetTypes
from parselmouth.exceptions import ParselmouthException
from parselmouth.targeting import AdUnit
from parselmouth.targeting import Custom
from parselmouth.targeting import Geography
from parselmouth.targeting import Placement
from parselmouth.targeting import TargetingCriterion
from parselmouth.targeting import TargetingData
from parselmouth.targeting import Technology


def clean_target_dict(data):
    """
    There are many unimplemented fields throughout parselmouth,
    for example some types of targeting are not yet implemented.
    In these cases, we must clean the dictionary before we pass
    it back to DFP. Fields like X.Type exist in many places, and these
    must be replaced with 'xsi_type' for exaample. Also, unneeded fields
    sometimes raise errors when passed back to DFP.

    @param data: dict
    @return: dict
    """
    if not isinstance(data, dict):
        return data

    # Track whether this is at bottom level of a nested dictionary
    at_bottom = True
    cln_dict = {}
    for _key, _val in data.items():
        if isinstance(_val, dict):
            at_bottom = False
            cln_dict[_key] = clean_target_dict(_val)
        elif isinstance(_val, list):
            at_bottom = False
            cln_dict[_key] = [
                clean_target_dict(d) for d in _val
            ]
        elif _key.find('_Type') >= 0:
            cln_dict['xsi_type'] = _val
        else:
            cln_dict[_key] = _val

    if at_bottom:
        # If you are at lowest level dictionary, only include
        # id and type fields
        clean_dict = {}
        for _key, _val in cln_dict.items():
            if _key.find('id') >= 0 or _key.find('Id') >= 0:
                clean_dict[_key] = _val
            elif _key == 'xsi_type':
                clean_dict[_key] = _val
    else:
        clean_dict = cln_dict

    return clean_dict


def _make_criterion_from_lists(includes, excludes):
    """
    Give a list of items to include and exclude,
    construct the corresponding TargetingCriterion

    @param includes: list
    @param excludes: list
    @return: TargetingCriterion|None
    """
    incl = TargetingCriterion(includes, TargetingCriterion.OPERATOR.OR)
    excl = TargetingCriterion(excludes, TargetingCriterion.OPERATOR.OR)
    if includes and excludes:
        return incl & (~excl)
    elif includes:
        return incl
    elif excludes:
        return ~excl
    else:
        return None


def _adunit_from_dfp(adunit):
    """
    Convert dfp dictionary to AdUnit targeting model

    @param adunit: dict
    @return: AdUnit
    """
    include_descendants = adunit.get('includeDescendants', False)

    return AdUnit(
        id=adunit['adUnitId'],
        include_descendants=include_descendants,
    )


def _adunit_to_dfp(adunit):
    """
    Convert AdUnit targeting model to dfp dictionary

    @param adunit: AdUnit
    @return: dict
    """
    return {
        'adUnitId': adunit.id,
        'includeDescendants': adunit.include_descendants,
    }


def transform_inventory_targeting_from_dfp(targeting):
    """
    Convert dfp dictionary to inventory targeting model

    @param targeting: dict
    @return: TargetingCriterion
    """
    if not targeting:
        return None

    include_list = []

    placement_ids = targeting.get('targetedPlacementIds', [])
    for pl_id in placement_ids:
        place = Placement(id=pl_id)
        include_list.append(place)

    ad_units = targeting.get('targetedAdUnits', [])
    for au in ad_units:
        include_list.append(_adunit_from_dfp(au))

    exclude_list = []
    ad_units = targeting.get('excludedAdUnits', [])
    for au in ad_units:
        exclude_list.append(_adunit_from_dfp(au))

    return _make_criterion_from_lists(include_list, exclude_list)


def transform_inventory_targeting_to_dfp(targeting):
    """
    Convert inventory targeting model to dfp dictionary

    @param adunit: TargetingCriterion
    @return: dict
    """
    targeted_placements = []
    targeted_adunits = []
    excluded_adunits = []

    includes, excludes = targeting.get_includes_and_excludes()

    for i in includes:
        if isinstance(i, AdUnit):
            targeted_adunits.append(_adunit_to_dfp(i))
        elif isinstance(i, Placement):
            targeted_placements.append(i.id)

    for e in excludes:
        if isinstance(e, AdUnit):
            excluded_adunits.append(_adunit_to_dfp(e))

    target_dict = {}
    if targeted_placements:
        target_dict['targetedPlacementIds'] = targeted_placements

    if targeted_adunits:
        target_dict['targetedAdUnits'] = targeted_adunits

    if excluded_adunits:
        target_dict['excludedAdUnits'] = excluded_adunits

    return target_dict


def _geo_from_dfp(geo):
    """
    Convert dfp dictionary to Geography targeting model

    @param adunit: dict
    @return: Geography
    """
    return Geography(
        id=geo['id'],
        type=geo['type'],
        name=geo['displayName'],
    )


def _geo_to_dfp(geo):
    """
    Convert Geography targeting model to dfp dictionary

    @param adunit: Geography
    @return: dict
    """
    return {
        'type': geo.type,
        'displayName': geo.name,
        'id': geo.id
    }


def transform_geography_targeting_from_dfp(targeting):
    """
    Convert dfp dictionary to geography targeting model

    @param targeting: dict
    @return: TargetingCriterion
    """
    if not targeting:
        return None

    exclude_list = []
    excluded = targeting.get('excludedLocations', [])
    for g in excluded:
        exclude_list.append(_geo_from_dfp(g))

    include_list = []
    included = targeting.get('targetedLocations', [])
    for g in included:
        include_list.append(_geo_from_dfp(g))

    return _make_criterion_from_lists(include_list, exclude_list)


def transform_geography_targeting_to_dfp(targeting):
    """
    Convert geography targeting model to dfp dictionary

    @param adunit: TargetingCriterion
    @return: dict
    """
    included_locations = []
    excluded_locations = []

    includes, excludes = targeting.get_includes_and_excludes()

    for i in includes:
        if isinstance(i, Geography):
            included_locations.append(_geo_to_dfp(i))

    for e in excludes:
        if isinstance(e, Geography):
            excluded_locations.append(_geo_to_dfp(e))

    target_dict = {}
    if included_locations:
        target_dict['targetedLocations'] = included_locations

    if excluded_locations:
        target_dict['excludedLocations'] = excluded_locations

    return target_dict


TECHNOLOGY_KEY_MAP = {
    TechnologyTargetTypes.bandwidth_group: {
        'target_name': 'BandwidthGroupTargeting',
        'type': 'list',
        'list_name': 'bandwidthGroups',
    },
    TechnologyTargetTypes.browser: {
        'target_name': 'browserTargeting',
        'type': 'list',
        'list_name': 'browsers',
    },
    TechnologyTargetTypes.browser_language: {
        'target_name': 'browserLanguageTargeting',
        'type': 'list',
        'list_name': 'browserLanguages',
    },
    TechnologyTargetTypes.device_manufacturer: {
        'target_name': 'DeviceManufacturerTargeting',
        'type': 'list',
        'list_name': 'deviceManufacturers',
    },
    TechnologyTargetTypes.mobile_carrier: {
        'target_name': 'MobileCarrierTargeting',
        'type': 'list',
        'list_name': 'mobileCarriers',
    },
    TechnologyTargetTypes.operating_system: {
        'target_name': 'OperatingSystemTargeting',
        'type': 'list',
        'list_name': 'operatingSystems',
    },
    TechnologyTargetTypes.device_capability: {
        'target_name': 'DeviceCapabilityTargeting',
        'type': 'categorical',
        'include': 'targetedDeviceCapabilities',
        'exclude': 'excludedDeviceCapabilities',
    },
    TechnologyTargetTypes.device_category: {
        'target_name': 'deviceCategoryTargeting',
        'type': 'categorical',
        'include': 'targetedDeviceCategories',
        'exclude': 'excludedDeviceCategories',
    },
    TechnologyTargetTypes.mobile_device: {
        'target_name': 'MobileDeviceTargeting',
        'type': 'categorical',
        'include': 'targetedMobileDevices',
        'exclude': 'excludedMobileDevices',
    },
    TechnologyTargetTypes.mobile_device_submodel: {
        'target_name': 'MobileDeviceSubmodelTargeting',
        'type': 'categorical',
        'include': 'targetedMobileDeviceSubmodels',
        'exclude': 'excludedMobileDeviceSubmodels',
    },
    TechnologyTargetTypes.operating_system_version: {
        'target_name': 'OperatingSystemVersionTargeting',
        'type': 'categorical',
        'include': 'targetedOperatingSystemVersions',
        'exclude': 'excludedOperatingSystemVersions',
    },
}
"""
dict, The way DFP handles technology targeting information
    is completely insane. This dictionary tells us how technology
    data must be handled for DFP to understand it properly.
    There are two different format types:
        1) list: A single list of ids is given, and a field
            isTargeted: True|False determines whether this list
            is targeted or not.
        2) categorical: Two lists of included/excluded ids
            are given.

    This dictionary also contains all the key names needed by DFP.
    For some reason they like to pluralize words in proper English
    in a way that cannot be easily automated.

    See: https://developers.google.com/doubleclick-publishers/docs/reference/v201408/ProposalLineItemService.TechnologyTargeting
"""


def transform_technology_targeting_from_dfp(targeting):
    """
    Convert DFP technology targeting data to TargetingCriterion

    @param targeting: dict
    @return: TargetingCriterion
    """
    if not targeting:
        return None

    target_data = {
        'include': [],
        'exclude': [],
    }

    for tech_type, _map in TECHNOLOGY_KEY_MAP.items():
        target = targeting.get(_map['target_name'])
        if not target:
            continue

        _map_type = _map['type']
        if _map_type == 'list':
            _list_name = _map['list_name']
            browser_list = []
            for item in target[_list_name]:
                tech_target = Technology(
                    id=item['id'],
                    name=item.get('name'),
                    type=tech_type,
                )
                browser_list.append(tech_target)
            # The isTargeted key determines whether
            # to target this entire list or not
            if target['isTargeted']:
                target_data['include'] += browser_list
            else:
                target_data['exclude'] += browser_list

        elif _map_type == 'categorical':
            # Get lists of items to be included and excluded
            for _cat_key in ['include', 'exclude']:
                dfp_key = _map[_cat_key]
                raw_data = target.get(dfp_key, [])
                for item in raw_data:
                    tech_target = Technology(
                        id=item['id'],
                        name=item.get('name'),
                        type=tech_type,
                    )
                    target_data[_cat_key].append(tech_target)

    # Build TargetingCriterion from list of includes and excludes
    includes = target_data['include']
    inc = TargetingCriterion(includes, TargetingCriterion.OPERATOR.OR)
    excludes = target_data['exclude']
    exl = TargetingCriterion(excludes, TargetingCriterion.OPERATOR.OR)
    if includes and excludes:
        return inc & (~exl)
    elif includes:
        return inc
    elif excludes:
        return ~exl
    else:
        return None


def transform_technology_targeting_to_dfp(targeting):
    """
    Convert TargetingCriterion to DFP technology targeting data
    in dictionary form.

    @param targeting: TargetingCriterion
    @return: dict
    """
    includes, excludes = targeting.get_includes_and_excludes()
    targeting_data = {
        'include': includes,
        'exclude': excludes,
    }

    dict_info = defaultdict(lambda: defaultdict(list))
    for criterion_type, criterion in targeting_data.iteritems():
        for item in criterion:
            tech_type = item.type

            _map = TECHNOLOGY_KEY_MAP[tech_type]
            _map_type = _map['type']
            _target_name = _map['target_name']

            # Only output the id of the technology target
            dfp_doc = {'id': item.id}
            if _map_type == 'list':
                _list_name = _map['list_name']
                dict_info[_target_name][_list_name].append(dfp_doc)
                dict_info[_target_name]['isTargeted'] = (criterion_type == 'include')
            elif _map_type == 'categorical':
                _category_name = _map[criterion_type]
                dict_info[_target_name][_category_name].append(dfp_doc)

    # Convert defaultdict to dictionary
    dfp_dict = {}
    for _key, _val in dict_info.items():
        dfp_dict[_key] = dict(_val)

    return dfp_dict


"""
The following targeting utils are unimplemented at this time.
If any of these fields need to be used within python,
these methods will need to be implemented.  For now, dictionaries
of native DFP structure will be stored into these targeting
fields to ensure that the data is not lost.
"""


def transform_user_domain_targeting_from_dfp(targeting):
    return clean_target_dict(targeting)


def transform_user_domain_targeting_to_dfp(targeting):
    return targeting or None


def transform_day_part_targeting_from_dfp(targeting):
    return clean_target_dict(targeting)


def transform_day_part_targeting_to_dfp(targeting):
    return targeting or None


def transform_video_content_targeting_from_dfp(targeting):
    return clean_target_dict(targeting)


def transform_video_content_targeting_to_dfp(targeting):
    return targeting or None


def transform_video_position_targeting_from_dfp(targeting):
    return clean_target_dict(targeting)


def transform_video_position_targeting_to_dfp(targeting):
    return targeting or None


def _recursive_custom_from_dfp(data):
    """
    Convert dfp dictionary to custom targeting model

    @param targeting: dict
    @return: TargetingCriterion
    """
    children = data.get('children', [])
    if children:
        operator = data['logicalOperator']
        target_list = [_recursive_custom_from_dfp(c) for c in children]
        return TargetingCriterion(target_list, operator)
    else:
        # Base case
        values = []

        if data.get('audienceSegmentIds'):
            node_key = 'AudienceSegmentCriteria'
            _id_key = 'audienceSegmentIds'
            value_ids = data[_id_key]
            for vid in value_ids:
                values.append(
                    Custom(
                        id=vid,
                        id_key=_id_key,
                        node_key=node_key,
                    )
                )
        else:
            node_key = 'CustomCriteria'
            _id_key = 'valueIds'
            value_ids = data[_id_key]
            key_id = data['keyId']
            for vid in value_ids:
                values.append(Custom(
                    id=vid,
                    id_key=_id_key,
                    parent_id=key_id,
                    node_key=node_key,
                ))

        targeting = TargetingCriterion(values, TargetingCriterion.OPERATOR.OR)
        if data['operator'] == 'IS':
            return targeting
        else:
            return ~targeting


def transform_custom_targeting_from_dfp(targeting):
    """
    Convert dfp dictionary to custom targeting model

    @param targeting: dict
    @return: TargetingCriterion
    """
    if not targeting:
        return None

    return _recursive_custom_from_dfp(targeting)


def _custom_target_list_to_child_list(custom_targets, is_operator):
    """
    Convert a list of custom targets to a list of dictionaries
    formatted for use in DFP

    @param custom_targets: list(Custom)
    @param is_operator: str, 'IS' or 'IS_NOT'
    @return: list(dict)
    """
    _children = defaultdict(lambda: defaultdict(list))
    for value in custom_targets:
        _children[value.parent_id][is_operator].append(value)

    dfp_children = []
    for _parent_id, _operations in _children.iteritems():
        for _operator, _values in _operations.iteritems():
            _value_target_ids = []
            _node_type = None
            _node_value_type = None

            for _custom_value in _values:
                _this_node_type = _custom_value.node_key
                _this_node_value_type = _custom_value.id_key

                if not _node_type and not _node_value_type:
                    _node_type = _this_node_type
                    _node_value_type = _this_node_value_type

                if (_this_node_type != _custom_value.node_key or
                        _this_node_value_type != _custom_value.id_key):
                    raise ParselmouthException(
                        "Could not serialize Custom Target: %s" % _custom_value
                    )
                _value_target_ids.append(_custom_value.id)

            _doc = {
                'operator': _operator,
                'xsi_type': _node_type,
                _node_value_type: _value_target_ids,
            }
            if _parent_id:
                _doc['keyId'] = _parent_id

            dfp_children.append(_doc)

    return dfp_children


def _recursive_custom_to_dfp(targeting):
    """
    Convert custom targeting model to dfp dictionary

    @param targeting: TargetingCriterion
    @return: dict
    """
    operator, targets = targeting.get_data()

    if operator == TargetingCriterion.OPERATOR.NOT:
        # In this case targets is of the form [TargetingCriterion]
        # NOTE: NOT operators are always a list of size 1
        # In this case we redefine operator and targets
        operator, targets = targets[0].get_data()
        is_operator = 'IS_NOT'
    else:
        is_operator = 'IS'

    if isinstance(targets[0], Custom):
        # In this case, we are in the deepest level of the targeting
        # targets is of type list(Custom)
        children = _custom_target_list_to_child_list(targets, is_operator)
    elif len(targets) == 1:
        # targets is a list of exactly one TargetingCriterion
        # To avoid redundant nesting, we return the output of this
        # individual result
        return _recursive_custom_to_dfp(targets[0])
    else:
        # In this case there are further depths to the targeting
        # targets is of type list(TargetingCriterion)
        children = [_recursive_custom_to_dfp(c) for c in targets]

    return {
        'xsi_type': 'CustomCriteriaSet',
        'children': children,
        'logicalOperator': operator
    }


def transform_custom_targeting_to_dfp(targeting):
    """
    Convert custom targeting model to dfp dictionary

    @param targeting: TargetingCriterion
    @return: dict
    """
    if not targeting:
        return None

    return _recursive_custom_to_dfp(targeting)


def transform_targeting_data_from_dfp(targeting):
    """
    Convert dictionary-representation for a SUDS creative object into a
    Parselmouth representation of a Creative

    @param targeting: dict, dictionary-representation of a DFP SUDS response
        for a single creative
    @return: parselmouth.delivery.TargetingData
    """

    return TargetingData(
        inventory=transform_inventory_targeting_from_dfp(targeting.get('inventoryTargeting')),
        geography=transform_geography_targeting_from_dfp(targeting.get('geoTargeting')),
        day_part=transform_day_part_targeting_from_dfp(targeting.get('dayPartTargeting')),
        user_domain=transform_user_domain_targeting_from_dfp(targeting.get('userDomainTargeting')),
        technology=transform_technology_targeting_from_dfp(targeting.get('technologyTargeting')),
        video_content=transform_video_content_targeting_from_dfp(targeting.get('contentTargeting')),
        video_position=transform_video_position_targeting_from_dfp(targeting.get('videoPositionTargeting')),
        custom=transform_custom_targeting_from_dfp(targeting.get('customTargeting')),
    )


def transform_targeting_data_to_dfp(targeting):
    """
    Convert dictionary-representation for a SUDS creative object into a
    Parselmouth representation of a Creative

    @param targeting: dict, dictionary-representation of a DFP SUDS response
        for a single creative
    @return: parselmouth.delivery.TargetingData
    """
    dfp_targeting = {}

    if targeting.inventory:
        dfp_targeting['inventoryTargeting'] = transform_inventory_targeting_to_dfp(targeting.inventory)

    if targeting.geography:
        dfp_targeting['geoTargeting'] = transform_geography_targeting_to_dfp(targeting.geography)

    if targeting.day_part:
        dfp_targeting['dayPartTargeting'] = transform_day_part_targeting_to_dfp(targeting.day_part)

    if targeting.technology:
        dfp_targeting['technologyTargeting'] = transform_technology_targeting_to_dfp(targeting.technology)

    if targeting.video_content:
        dfp_targeting['contentTargeting'] = transform_video_content_targeting_to_dfp(targeting.video_content)

    if targeting.video_position:
        dfp_targeting['videoPositionTargeting'] = transform_video_position_targeting_to_dfp(targeting.video_position)

    if targeting.custom:
        dfp_targeting['customTargeting'] = transform_custom_targeting_to_dfp(targeting.custom)

    return dfp_targeting




#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - Configuration Utilities

These configuration utilities outline the way configurations may be
loaded into Parselmouth for authenticating different advertising client
APIs
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
from abc import ABCMeta
from abc import abstractmethod

# Third Party Imports
import yaml

# Parselmouth Imports
from parselmouth.exceptions import ParselmouthException
from parselmouth.constants import ParselmouthProviders

# Parselmouth Imports - Adapter Imports
from parselmouth.adapters.dfp.config import DFPConfig


class ParselmouthConfig(object):
    """
    Abstract Configuration Container

    This class outlines an interface for loading provider-specific
    configurations. These classes can be extended to allow for alternate
    data sources such as an external credential database instead of a
    file for instance.
    """

    ProviderConfigInterfaceMap = {
        ParselmouthProviders.google_dfp_premium: DFPConfig,
        ParselmouthProviders.google_dfp_small_business: DFPConfig,
    }
    """
    dict, mapping between ad service providers and their configuration
    containers
    NOTES:
        * This doesn't serve much purpose right now since all we support
            is DFP...
    """

    def __init__(self,
                 provider_name,
                 config_path=None,
                 **kwargs):
        """
        @param provider_name: ParselmouthProviders
        """
        self.provider_name = provider_name
        self.provider_config = self.get_config_for_provider(
            self.provider_name,
        )

        if config_path:
            try:
                self.provider_config.load_config_from_file(
                    self.provider_name, config_path,
                )
            except:
                raise ParselmouthException("Invalid credential file.")
        else:
            try:
                self.provider_config.load_config_from_external_source(**kwargs)
            except Exception:
                raise ParselmouthException("Invalid credential arguments.")

        self.provider_config.validate_credentials()

    def __str__(self):
        """
        Human readable representation of this object
        @return: str
        """
        return "{class_name}({vars})".format(
            class_name=self.__class__.__name__,
            vars=str(vars(self))
        )

    @classmethod
    def get_config_for_provider(cls, provider_name):
        """
        Staticmethod, returns the interface class for a given ad service
        provider name

        @param provider_name: str, one of enum
            parselmouth.constants.ParselmouthProviders
        @return: descendant of
            parselmouth.adapters.abstract_config.AbstractConfig
        """
        client_config = cls.ProviderConfigInterfaceMap.get(provider_name)
        if not client_config:
            raise ValueError(
                "There is no config interface defined for provider: %s" %
                provider_name
            )
        return client_config()

    def get_credentials_arguments(self):
        """
        @return: dict, dictionary of all values needed to connect
            to the associated ad provider service
        """
        return self.provider_config.get_credentials_arguments()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - Exceptions
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class ParselmouthException(Exception):
    """ Base Exception for the Parselmouth project
    """
    pass

class ParselmouthNetworkError(Exception):
    """ Base Exception for the Parselmouth Network Errors
    """
    pass

class ParselmouthTimeout(ParselmouthNetworkError):
    """
    Raised when the block under context management takes longer to
    complete than the allowed maximum timeout value.
    """
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" DFP Native Client
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
import csv
import logging
from gzip import GzipFile
from tempfile import NamedTemporaryFile

# Third Party Library Imports
from googleads import DfpClient as GoogleDFPClient
from googleads.oauth2 import GoogleRefreshTokenClient
from googleads.dfp import FilterStatement
from googleads.dfp import SUGGESTED_PAGE_LIMIT
from googleads.errors import DfpReportError

# Parselmouth Imports
from parselmouth.exceptions import ParselmouthException

# Parselmouth Imports - Local DFP Adapter Imports
from parselmouth.adapters.dfp.constants import DFP_API_VERSION
from parselmouth.adapters.dfp.constants import DFP_CUSTOM_TARGETING_KEY_TYPES
from parselmouth.adapters.dfp.constants import DFP_QUERY_DEFAULTS
from parselmouth.adapters.dfp.constants import DFP_VALUE_MATCH_TYPES
from parselmouth.adapters.dfp.utils import format_pql_response
from parselmouth.adapters.dfp.utils import format_report_list
from parselmouth.adapters.dfp.utils import sanitize_report_response


class DFPClient(object):
    """
    Class retrieving and submitting data to the DFP API

    This is meant to be a pretty bare-bones class for simply building
    queries submittable to DFP.
    """

    def __init__(self,
                 client_id,
                 client_secret,
                 refresh_token,
                 application_name,
                 network_code,
                 version=DFP_API_VERSION):
        """
        https://developers.google.com/doubleclick-publishers/docs/authentication

        @param client_id: str
        @param client_secret: str
        @param refresh_token: str
        @param application_name: str
        @param network_code: str
        @param version: str
        """
        self.version = DFP_API_VERSION
        self.native_dfp_client = self._get_client(
            client_id,
            client_secret,
            refresh_token,
            application_name,
            network_code,
        )

    def __repr__(self):
        """
        Human readable representation of this object

        @return: str
        """
        return (
            "{class_name}("
                "version='{version}'"
            ")"
        ).format(
            class_name=self.__class__.__name__,
            version=self.version
        )

    def _get_client(self,
                    client_id,
                    client_secret,
                    refresh_token,
                    application_name,
                    network_code):
        """
        Create a client to connect the Google DFP API

        @param headers: dict, Oauth2 authorization header
        @return: googleads.DfpClient
        """
        client_token = GoogleRefreshTokenClient(
            client_id,
            client_secret,
            refresh_token,
        )
        return GoogleDFPClient(
            client_token,
            application_name,
            network_code,
        )

    def _format_query(self,
                      order=DFP_QUERY_DEFAULTS['order'],
                      limit=None,
                      offset=DFP_QUERY_DEFAULTS['offset'],
                      **filter_kwargs):
        """
        Query formatter

        @param query_dict: dictionary of query parameters to organize
            into a PQL statemente
        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: FilterStatement, PQL statement

        ## TODO: We probably want to do some checking of `filter_kwargs`
            against available column names in DFP PQL tables
        """
        if order:
            query = "ORDER BY {order}".format(order=order)
        else:
            query = ""

        filters = []
        for key, val in filter_kwargs.iteritems():
            if isinstance(val, list):
                # gotta format that in query
                _dfp_list = ", ".join([str(x) for x in val])
                filters.append("{0} IN ({1})".format(key, _dfp_list))
                pass
            else:
                filters.append("{0}={1}".format(key, val))
        if filters:
            # Prepend the where parameters to the base query
            where_query = " AND ".join(filters)

            query = "WHERE {where} {order}".format(
                where=where_query,
                order=query,
            )

        statement = FilterStatement(query)
        if limit:
            statement.limit = limit
        if offset:
            statement.offset = offset

        return statement

    def _run_service_query(self, query, query_function):
        """
        Run a series of chunked DFP queries until all results
        are acquired

        @param query: FilterStatement
        @param query_function: Dfp service method
        @return: list
        """
        results = []
        while True:
            try:
                response = query_function(query.ToStatement())
            except Exception as e:
                raise ParselmouthException(
                    "Error running query: {0}. Got Error: {1}".format(
                        query.ToStatement(),
                        str(e)
                    )
                )
            if 'results' in response:
                logging.info(
                    'Statement %s returned %d results',
                    query.ToStatement(),
                    len(response['results']),
                )
                results += response['results']
                query.offset += SUGGESTED_PAGE_LIMIT
            else:
                break

        return results

    def get_network_data(self):
        """
        Get network data associated with dfp account

        @return: SUDS envelope
        """
        service = self.native_dfp_client.GetService(
            'NetworkService',
            version=self.version,
        )
        return service.getCurrentNetwork()

    def get_order(self, order_id):
        """
        Gets an order item by id

        @param order_id: str|int, the numerical id of the item
        @return: SUDS envelope
        """
        return self.get_orders(id=order_id)

    def get_orders(self,
                   order=DFP_QUERY_DEFAULTS['order'],
                   limit=DFP_QUERY_DEFAULTS['limit'],
                   offset=DFP_QUERY_DEFAULTS['offset'],
                   **filter_kwargs):
        """
        Get orders on filter keyword arguments other than id

        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: SUDS envelope

        Example:
            * Get an order by id: `get_orders(id=ORDER_ID)`
        """
        query = self._format_query(
            order=order,
            limit=limit,
            offset=offset,
            **filter_kwargs
        )
        service = self.native_dfp_client.GetService(
            'OrderService',
            version=self.version
        )

        orders = self._run_service_query(
            query, service.getOrdersByStatement,
        )

        if not orders:
            logging.warning(
                'Results not found from query. Query: {0}'.format(query.ToStatement())
            )
            return []

        return orders

    def get_line_item(self, line_item_id):
        """
        Gets a line item by id

        @param item_id: str|int, the numerical id of the item
        @return: SUDS envelope
        """
        return self.get_line_items(id=line_item_id)

    def get_line_items(self,
                       order=DFP_QUERY_DEFAULTS['order'],
                       limit=DFP_QUERY_DEFAULTS['limit'],
                       offset=DFP_QUERY_DEFAULTS['offset'],
                       **filter_kwargs):
        """
        Get line items on filter keyword arguments other than id

        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: SUDS envelope

        Example:
            * Get an order by id: `get_line_items(id=LINE_ITEM_ID)`
        """
        query = self._format_query(
            order=order,
            limit=limit,
            offset=offset,
            **filter_kwargs
        )
        service = self.native_dfp_client.GetService(
            'LineItemService',
            version=self.version
        )

        line_items = self._run_service_query(
            query, service.getLineItemsByStatement,
        )

        if not line_items:
            logging.warning(
                'Results not found from query. Query: {0}'.format(query)
            )
            return []

        return line_items

    def get_advertisers(self):
        """
        Queries dfp for all advertisers within their account

        @return: list(dict), returns a list of company dictionaries
        """
        service = self.native_dfp_client.GetService(
            'CompanyService',
            version=self.version,
        )

        # Create statement object to only select companies that are advertisers
        values = [{
            'key': 'type',
            'value': {
                'xsi_type': 'TextValue',
                'value': 'ADVERTISER'

            }
        }]
        query = 'WHERE type = :type'
        statement = FilterStatement(query, values)
        results = self._run_service_query(statement, service.getCompaniesByStatement)
        advertisers = [
            {'id': advertiser['id'],
             'name': advertiser['name']
             }
            for advertiser in results
        ]
        return advertisers

    def forecast_line_item(self, dfp_line_item):
        """
        @param line_item: LineItem
        @return: SUDS envelope
        """
        service = self.native_dfp_client.GetService(
            'ForecastService',
            version=self.version,
        )
        prospective_line_item = {
            'lineItem': dfp_line_item,
        }
        forecast_options = {
            'includeContendingLineItems': False,
            'includeTargetingCriteriaBreakdown': False,
        }
        return service.getAvailabilityForecast(
            prospective_line_item,
            forecast_options,
        )

    def get_creative(self, creative_id):
        """
        Gets a creative by id

        @param item_id: str|int, the numerical id of the item
        @return: SUDS envelope
        """
        return self.get_creatives(id=creative_id)

    def get_creatives(self,
                      order=DFP_QUERY_DEFAULTS['order'],
                      limit=DFP_QUERY_DEFAULTS['limit'],
                      offset=DFP_QUERY_DEFAULTS['offset'],
                      **filter_kwargs):
        """
        Get creatives on filter keyword arguments other than id

        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: SUDS envelope

        Example:
            * Get an order by id: `get_creatives(id=CREATIVE_ID)`
        """
        query = self._format_query(
            order=order,
            limit=limit,
            offset=offset,
            **filter_kwargs
        )
        service = self.native_dfp_client.GetService(
            'CreativeService',
            version=self.version
        )

        creatives = self._run_service_query(
            query, service.getCreativesByStatement,
        )

        if not creatives:
            logging.warning(
                'Results not found from query. Query: {0}'.format(query)
            )
            return []

        return creatives

    def get_line_item_creatives(self, line_item_id):
        """
        Get the creatives assoicated with a particular line item

        @param line_item_id: int, the line item id
        @return: list(str), list of creative ids associated with the
            given line item
        """
        values = [{
              'key': 'lineItemId',
              'value': {
                  'xsi_type': 'NumberValue',
                  'value': line_item_id
              }
          }]
        query = 'WHERE lineItemId = :lineItemId'
        statement = FilterStatement(where_clause=query, values=values)
        service = self.native_dfp_client.GetService(
            'LineItemCreativeAssociationService',
            version=self.version
        )

        creatives = self._run_service_query(
            statement, service.getLineItemCreativeAssociationsByStatement,
        )

        if creatives:
            if 'results' in creatives:
                return [
                    str(creative['creativeId'])
                    for creative in creatives['results']
                ]
            else:
                return [
                    str(creative['creativeId'])
                    for creative in creatives
                ]
        else:
            return []

    def _run_pql_query(self, pql_query, values=None):
        """
        Get a list of contents of a DFP table for the specific pql query

        @param pql_query: str
        @param values: list(dict)
        @return: list
        """
        report_downloader = self.native_dfp_client.GetDataDownloader(
            version=self.version,
        )
        raw_list = report_downloader.DownloadPqlResultToList(
            pql_query, values,
        )
        return format_pql_response(raw_list)

    def get_geography_targets(self):
        """
        Get a list of DFP geography ids

        @return: list(dict)
        """
        pql_query = """
        SELECT Id, Name, CountryCode, Type
        FROM Geo_Target
        WHERE type='country'
            AND targetable=true
        """
        return self._run_pql_query(pql_query)

    def get_adunit_targets(self):
        """
        Get a list of DFP geography ids

        @return: list(dict)
        """
        pql_query = """
        SELECT Id, Name, ParentId
        FROM Ad_Unit
        """
        return self._run_pql_query(pql_query)

    def get_custom_targets(self, key_name=None, value_name=None):
        """
        Get all custom targeting key and value data

        @param key_name: str|None, if present only return
            targeting with the given key name
        @param value_name: str|None, if present only return
            targeting with the given value name
        @return: list(dict)
        """
        service = self.native_dfp_client.GetService(
            'CustomTargetingService',
            version=self.version,
        )

        if key_name:
            values = [{
                'key': 'name',
                'value': {
                    'xsi_type': 'TextValue',
                    'value': key_name,
                }
            }]
            query = 'WHERE name = :name'
            key_statement = FilterStatement(query, values)
        else:
            key_statement = FilterStatement()

        key_results = self._run_service_query(
            key_statement, service.getCustomTargetingKeysByStatement,
        )

        custom_data = []
        if not key_results:
            return custom_data

        if not value_name:
            # In the case when value_name is given, do not include
            # the key target
            custom_data += key_results

        # Create statement to get all targeting values.
        query = 'WHERE customTargetingKeyId IN ({})'.format(
            ', '.join([str(key['id']) for key in key_results]),
        )

        if value_name:
            values = [{
                'key': 'name',
                'value': {
                    'xsi_type': 'TextValue',
                    'value': value_name,
                }
            }]
            query += ' AND name = :name'
            value_statement = FilterStatement(query, values)
        else:
            value_statement = FilterStatement(query)

        # Get custom targeting values by statement.
        value_results = self._run_service_query(
            value_statement, service.getCustomTargetingValuesByStatement,
        )
        custom_data += value_results
        return custom_data

    def create_custom_target(self,
                             key_name,
                             key_type,
                             value_name,
                             key_display_name=None,
                             value_display_name=None,
                             value_match_type=None):
        """
        Add a custom target to DFP

        NOTE: Since DFP is responsible for creating the custom target
            key and value ids we must first create the key to get its id
            before creating the value

        DFP custom targeting keys are paraphrased here:
            {
                id: int- id of custom targeting value (assigned by DFP)
                name: str- name of the value, can be used for encoding
                displayName: str|None- descriptive name of the value
                type: ENUM(PREDEFINED, FREEFORM), see docs
            }
        The full spec can be found here:
            https://developers.google.com/doubleclick-publishers/docs/reference/v201408/CustomTargetingService.CustomTargetingKey

        DFP custom targeting values are paraphrased here:
            {
                customTargetingKeyId: int- id of custom targeting key
                id: int- id of custom targeting value (assigned by DFP)
                name: str- name of the value, can be used for encoding
                displayName: str|None- descriptive name of the value
                matchType: ENUM(EXACT, BROAD, PREFIX, BROAD_PREFIX,
                           SUFFIX, CONTAINS, UNKNOWN)|None, see docs
            }
        The full outline can be found here:
            https://developers.google.com/doubleclick-publishers/docs/reference/v201408/CustomTargetingService.CustomTargetingValue

        @param key_name: str, encoded name for the key
        @param key_type: str, type of the custom targeting key
        @param value_name: str, encoded name for the value
        @param key_display_name: str|None, descriptive name for key
        @param value_display_name: str|None, descriptive name for value
        @return: list(dict)
        """

        # Input checking
        if key_type not in DFP_CUSTOM_TARGETING_KEY_TYPES:
            raise ParselmouthException(
                "Provided key type ({0}) not one of a valid type ({1})".format(
                    key_type, DFP_CUSTOM_TARGETING_KEY_TYPES
                ))
        if value_match_type and value_match_type not in DFP_VALUE_MATCH_TYPES:
            raise ParselmouthException(
                "Provided val type ({0}) not one of a valid type ({1})".format(
                    value_match_type, DFP_VALUE_MATCH_TYPES
                ))

        existing_values = self.get_custom_targets(
            key_name=key_name,
            value_name=value_name
        )

        # Return if we already have a custom targeting keypair for key:value
        if existing_values:
            logging.info("Custom target key+value already exists in DFP")
            return existing_values

        service = self.native_dfp_client.GetService(
            'CustomTargetingService',
            version=self.version,
        )

        # Check if the key already exists in DFP, if not then create it
        # We need to do this first since DFP handles id assignment
        existing_keys = self.get_custom_targets(key_name=key_name)
        if not existing_keys:
            logging.info("Key ({0}) does not exist in DFP".format(key_name))
            key = {
                'name': key_name,
                'type': key_type,
            }
            if key_display_name:
                key['displayName'] = key_display_name
            keys = service.createCustomTargetingKeys([key])
        else:
            logging.info("Key ({0}) already exists in DFP".format(key_name))
            keys = existing_keys

        # Extract the key we want
        if len(keys) == 0:
            raise ParselmouthException("No keys returned by DFP")
        elif len(keys) > 1:
            raise ParselmouthException("Too many keys returned by DFP")
        key = keys[0]

        # Create the new value in DFP
        logging.info("Value ({0}) does not exist in DFP".format(value_name))
        value = {
            'customTargetingKeyId': key['id'],
            'name': value_name,
        }
        if value_display_name:
            value['displayName'] = value_display_name
        if value_match_type:
            value['matchType'] = value_match_type
        values = service.createCustomTargetingValues([value])

        # Extract the value we want
        if len(values) == 0:
            raise ParselmouthException("No values returned by DFP")
        elif len(values) > 1:
            raise ParselmouthException("Too many values returned by DFP")
        value = values[0]

        if value['customTargetingKeyId'] != key['id']:
            raise ParselmouthException((
                "Value custom target key id ({0}) does not equal the Key id "
                "({1})".format(value['customTargetingKeyId'], key['id'])
            ))

        return [key, value]

    def _generate_report_as_list(self, report_query):
        """
        Generates a report from a report query.  This
        function will hang until the report has finished processing.
        This function does not return the report, but a pointer
        to this report

        @param report_query: dict
        @return: list(dict)
        """
        logging.info('Generating report with query: %s', report_query)
        report_downloader = self.native_dfp_client.GetDataDownloader(
            version=self.version,
        )
        try:
            # Run the report and wait for it to finish.
            report_id = report_downloader.WaitForReport(
                {'reportQuery': report_query},
            )
        except DfpReportError, e:
            logging.exception(e)
            return None

        with NamedTemporaryFile(suffix='.csv.gz', delete=True) as report_file:
            # Download report data.
            report_downloader.DownloadReportToFile(
                report_id, 'CSV_DUMP', report_file
            )
            # Go to top of file
            report_file.seek(0)
            # Unzip contents and read
            with GzipFile(fileobj=report_file, mode='r') as unzipped:
                csvfile = csv.reader(unzipped)
                parsed_data = [
                    [cell.decode('utf-8') for cell in row]
                    for row in csvfile
                ]

        logging.info('Report download completed with %d results', len(parsed_data))
        return format_report_list(parsed_data)

    def generate_report(self,
                        dimensions,
                        columns,
                        date_range_type,
                        start_date=None,
                        end_date=None,
                        filter_query=None):
        """
        https://developers.google.com/doubleclick-publishers/docs/reference/v201408/ReportService.ReportQuery

        @param dimensions: list(str), list of DFP dimensions
        @param columns: list(str), list of DFP columns
        @param date_range_type: str, e.g. LAST_WEEK, CUSTOM_DATE
        @param start_date: datetime|None, if date_range_type is CUSTOM_DATE, this is required
        @param end_date: datetime|None, if date_range_type is CUSTOM_DATE, this is required
        @param filter_statement: str|None, filter to apply to report query
        @return: list(dict)
        """
        report_query = {
            'dimensions': dimensions,
            'columns': columns,
            'dateRangeType': date_range_type,
        }

        if date_range_type == 'CUSTOM_DATE':
            assert start_date and end_date
            report_query['startDate'] = {
                'year': start_date.year,
                'month': start_date.month,
                'day': start_date.day,
            }
            report_query['endDate'] = {
                'year': end_date.year,
                'month': end_date.month,
                'day': end_date.day,
            }

        if filter_query:
            report_query['statement'] = filter_query

        raw_list = self._generate_report_as_list(report_query)
        return [sanitize_report_response(i) for i in raw_list]

    def update_line_items(self, line_items):
        """
        Update a list of line items in DFP

        @param line_items: L{dict}
        """

        service = self.native_dfp_client.GetService(
            'LineItemService',
            version=self.version
        )
        service.updateLineItems(line_items)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth Abstract Interface
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
from abc import ABCMeta
from abc import abstractmethod


class AbstractInterface(object):
    """
    Abstract Interface to all Ad Services

    This class outlines functions required to be implemented by ad
    services. If that functionality is not provided by a specific ad
    service then it needs to raise `NotImplementedError`.

    Right now most of this functionality follows the DFP implementation.
    It's likely that many of these functions will not be used by other
    ad services.
    """
    __metaclass__ = ABCMeta

    def __repr__(self):
        """
        Human readable representation of this object

        @return: str
        """
        return (
            "{class_name}("
            ")"
        ).format(
            class_name=self.__class__.__name__,
        )

    @abstractmethod
    def get_network_timezone(self):
        pass

    @abstractmethod
    def get_campaign(self, creative_id):
        pass

    @abstractmethod
    def get_campaigns(self):
        pass

    @abstractmethod
    def get_line_item(self, line_item_id):
        pass

    @abstractmethod
    def get_line_items(self):
        pass

    @abstractmethod
    def get_campaign_line_items(self, campaign_id):
        pass

    @abstractmethod
    def get_line_item_available_inventory(self, line_item):
        pass

    @abstractmethod
    def get_advertisers(self, line_item):
        pass

    @abstractmethod
    def get_creative(self, creative_id):
        pass

    @abstractmethod
    def get_creatives(self):
        pass

    @abstractmethod
    def get_line_item_creatives(self, line_item_id):
        pass

    @abstractmethod
    def get_geography_targets(self):
        pass

    @abstractmethod
    def get_adunit_targets(self):
        pass

    @abstractmethod
    def get_custom_targets(self):
        pass

    @abstractmethod
    def get_line_item_report(self):
        pass

    @abstractmethod
    def update_line_items(self):
        pass

    @abstractmethod
    def create_custom_target(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth

This package serves as the base interface from third-party ad servers to
python object-oriented classes. This file exposes only the objects
clients should use.

Parselmouth should be imported in the following way:
    `from parselmouth import Parselmouth`
"""

# Expose package interfaces
__all__ = [
    'Parselmouth',
    'ParselmouthConfig',
    'ParselmouthException',
    'ParselmouthProviders',
]

from parselmouth.base import Parselmouth
from parselmouth.config import ParselmouthConfig
from parselmouth.exceptions import ParselmouthException
from parselmouth.constants import ParselmouthProviders



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Parselmouth utilities - check_equal
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


def check_equal(obj1, obj2):
    """
    Check if two objects are equivalent

    @param obj1
    @parma obj2
    @return: bool
    """
    if type(obj1) != type(obj2):
        return False

    if isinstance(obj1, dict):
        if sorted(obj1.keys()) != sorted(obj2.keys()):
            return False

        for _key in obj1.keys():
            if not check_equal(obj1[_key], obj2[_key]):
                return False

    elif isinstance(obj1, list):
        if len(obj1) != len(obj2):
            return False

        for item1 in obj1:
            is_match = False
            for item2 in obj2:
                if check_equal(item1, item2):
                    is_match = True

            if not is_match:
                return False
    else:
        return obj1 == obj2

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - Tree Builder

Ad providers such as DFP employ complex tree structures to organize
zones or adunits within their system.  The TreeBuilder class helps
build and manipulate collections of objects with these tree structures.
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
from collections import defaultdict

# Parselmouth Imports
from parselmouth.constants import ParselmouthTargetTypes
from parselmouth.targeting import AdUnit
from parselmouth.targeting import Custom
from parselmouth.targeting import Geography


class NodeTree(object):
    """
    Class which represents a tree of ObjectModels
    """
    def __init__(self, node, children, depth=None):
        self.node = node
        self.children = children
        self.depth = depth

    def __ne__(self, other):
        return not(self == other)

    def __eq__(self, other):
        if self.node != other.node or self.depth != other.depth:
            return False

        if len(self.children) != len(other.children):
            return False

        for child in self.children:
            matching_child = [c for c in other.children if c == child]
            if not matching_child:
                return False

        return True

    def to_doc(self):
        return {
            'node': self.node.to_doc() if self.node else None,
            'children': [c.to_doc() for c in self.children],
            'depth': self.depth,
        }

    def get_subtree(self, field_name, field_value):
        """
        Find a subtree within a list of NodeTrees with the field value given

        @param field_name: ParselmouthFields
        @param field_value: str
        @return: NodeTree
        """
        subtree = None
        if self.children:
            for branch in self.children:
                _node = branch.node
                if _node and vars(_node).get(field_name) == field_value:
                    subtree = branch
                elif branch.children:
                    subtree = branch.get_subtree(
                        field_name,
                        field_value,
                    )

                if subtree:
                    return subtree

        return subtree

    def get_subtree_parents(self, field_name, field_value):
        """
        Get the node associated to the given field_name/field_value,
        then return a list of all parent nodes for this node
        =
        @param field_name: ParselmouthFields
        @param field_values: str
        @return: list(ObjectModel)
        """
        parent_nodes = []
        if self.node and self.get_subtree(field_name, field_value):
            parent_nodes.append(self.node)

        if self.children:
            for branch in self.children:
                parent_nodes += branch.get_subtree_parents(field_name, field_value)

        return parent_nodes

    def get_max_depth(self):
        """
        Get the maximum depth of any node within the tree

        @return: int|None
        """
        max_depth = self.depth

        if self.children:
            for branch in self.children:
                max_depth = max([max_depth, branch.get_max_depth()])

        return max_depth

    def flatten(self, depth=None, only_leaves=False):
        """
        Get a flat list of all descendants from a subtree.
        If depth is given, give only nodes at the given depth

        @param depth: int|None
        @param only_leaves: bool, only return maximal depth nodes
        @return: list(ObjectModel)
        """
        if only_leaves:
            assert depth is None

        descendants = []
        if self.node and \
            (depth is None or self.depth == depth) and \
            (only_leaves is False or len(self.children) == 0):
            descendants.append(self.node)

        if self.children:
            for branch in self.children:
                if branch.children:
                    descendants += branch.flatten(
                        depth=depth,
                        only_leaves=only_leaves,
                    )
                elif branch.node and (depth is None or branch.depth == depth):
                    descendants.append(branch.node)

        return descendants

    def filter_tree_by_key(self, key, filter_ids):
        """
        Filter a given tree to include branches that are either
        included in the set of filter_ids or at least of of its
        children are in the set of filter_ids

        @param key: ParselmouthField, key to filter on
        @param filter_ids: set(str)
        @return: NodeTree
        """
        if isinstance(filter_ids, list):
            filter_set = set(filter_ids)
        else:
            filter_set = filter_ids

        assert isinstance(filter_set, set)

        filtered_children = []
        for branch in self.children:
            _descendants = branch.flatten()
            _descendant_ids = set([vars(d)[key] for d in _descendants])

            # if some descendant of this branch is in filter_ids
            # keep this branch and update its children
            if filter_set.intersection(_descendant_ids):
                new_branch = branch.filter_tree_by_key(
                    key, filter_set,
                )
                if new_branch:
                    filtered_children.append(new_branch)

        return NodeTree(self.node, filtered_children, self.depth)

    def filter_tree(self, filter_ids):
        """
        Filter a given tree to include branches that are either
        included in the set of filter_ids or at least of of its
        children are in the set of filter_ids

        @param tree: list(dict)
        @param filter_ids: set(str)
        @return: NodeTree
        """
        return self.filter_tree_by_key('external_name', filter_ids)

    def update_external_names(self, id_map):
        """
        Set external names of nodes in given tree to new values
        given by the dictionary id_map: id field --> external_name field.
        NOTE: This edits the nodes of this tree in place

        @param id_map: dict
        """
        current_node = self.node
        if current_node:
            _external_name = id_map.get(current_node.id)
            if _external_name:
                current_node.external_name = _external_name

        for branch in self.children:
            branch.update_external_names(id_map)


class TreeBuilder(object):
    """
    Interface from building ad provider data into trees

    Subject include:
        * Demographic data
        * Geographic data
        * Ad Placement data
    """

    TARGET_CLASS_MAP = {
        ParselmouthTargetTypes.adunit: AdUnit,
        ParselmouthTargetTypes.geography: Geography,
        ParselmouthTargetTypes.demographics: Custom,
        ParselmouthTargetTypes.ad_position: Custom,
        ParselmouthTargetTypes.custom: Custom,
    }
    """
    dict, associate to each target_type the appropriate targeting class
    """

    INTERFACE_FUNCTION_MAP = {
        ParselmouthTargetTypes.adunit: lambda i: i.get_adunit_targets(),
        ParselmouthTargetTypes.geography: lambda i: i.get_geography_targets(),
        ParselmouthTargetTypes.demographics: lambda i: i.get_custom_targets(),
        ParselmouthTargetTypes.ad_position: lambda i: i.get_custom_targets(),
        ParselmouthTargetTypes.custom: lambda i: i.get_custom_targets(),
    }
    """
    dict, associate to each target_type the interface target getter function
    """

    def __init__(self, provider_name, interface=None):
        """
        Constructor

        @param domain: str
        @param provider_name: ParselmouthProviders
        @param interface: Interface for provider
        """
        self.provider_name = provider_name
        self.interface = interface

    def _recursive_make_tree(self, pid, parents, node_map, depth=0):
        """
        Recursively construct tree by creating a nested dictionary
        based off of data from parents that gives a parent child
        relationship

        @param pid: parent id
        @param parents: dict, the keys are parent ids, and values
            are all of the children of that id
        @param node_map: dict, node documents associated with each id
        @param depth: int, recursive handle on node depth in a tree
        @return: list(dict)
        """

        trees = []
        for child in parents.get(pid, []):
            _id = child.id
            node = node_map.get(_id)
            tree = NodeTree(
                node,
                children=self._recursive_make_tree(
                    _id, parents, node_map, depth + 1
                ),
                depth=depth,
            )
            trees.append(tree)

        return trees

    def build_tree(self, nodes):
        """
        Interface for translating a flat data structure into a native
        hierarchical type.

        @param nodes: list(TargetingModel) with id and parent_id fields
        @return: NodeTree
        """
        node_map = {}
        # Create flat list of all parents and their immediate children
        parents = defaultdict(list)
        for node in nodes:
            parents[node.parent_id].append(node)
            node_map[node.id] = node

        # Construct tree by building trees at each maximal parent
        maximal_trees = []
        for pid in parents.keys():
            # Determine if pid is a maximal parent
            # By checking to see if pid is a child of any other id
            maximal = True
            for children in parents.values():
                children_ids = [c.id for c in children]
                # If pid is a child of another id, then flag as not maximal
                if pid in children_ids:
                    maximal = False
                    break
            # Create tree starting at this maximal parent
            if maximal:
                max_node = node_map.get(pid)
                children = self._recursive_make_tree(
                    pid, parents, node_map,
                )

                if max_node:
                    tree = NodeTree(max_node, children)
                    maximal_trees.append(tree)
                else:
                    maximal_trees.extend(children)

        return NodeTree(
            node=None,
            children=maximal_trees,
        )

    def construct_tree(self, target_type):
        """
        Get all data of type target_type from ad provider,
        and build a tree

        @param taget_type: ParselmouthTargetTypes
        @return: NodeTree
        """
        assert target_type in ParselmouthTargetTypes
        assert self.interface
        nodes = self.INTERFACE_FUNCTION_MAP[target_type](self.interface)
        return self.build_tree(nodes)

    def _convert_node_tree_to_doc(self, tree):
        """
        Convert NodeTree into a dict that can be written to mongo

        @param tree: NodeTree
        @return: dict
        """
        return tree.to_doc()

    def _convert_doc_to_node_tree(self, doc, target_type):
        """
        Convert a mongo tree doc into a NodeTree Object

        @param doc: dict
        @param target_type: ParselmouthTargetTypes
        @return: NodeTree
        """
        _target_class = self.TARGET_CLASS_MAP[target_type]
        if doc['node']:
            node = _target_class.from_doc(doc['node'])
        else:
            node = None

        children = doc.get('children', [])
        depth = doc.get('depth')

        return NodeTree(
            node=node,
            children=[self._convert_doc_to_node_tree(c, target_type) for c in children],
            depth=depth,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - Targeting Models

Base Interfaces for targeting items
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
from copy import deepcopy

# Parselmouth Imports
from parselmouth.exceptions import ParselmouthException
from parselmouth.model import ObjectModel
from parselmouth.utils.enum import Enum
from parselmouth.utils.check import check_equal


class AdUnit(ObjectModel):
    """
    Abstract Class for Ad Units

    An ad unit is a representation of one or more spaces where ads can
    be delivered.
    """
    def __init__(self,
                 id=None,
                 parent_id=None,
                 name=None,
                 include_descendants=True,
                 external_id=None,
                 external_name=None):
        self.id = id
        self.parent_id = parent_id
        self.name = name
        self.include_descendants = include_descendants
        self.external_id = external_id
        self.external_name = external_name


class Placement(ObjectModel):
    """
    Abstract Class for Placements

    A placement is an optional grouping of ad units that make them
    easier to target all at once.
    """
    def __init__(self,
                 id=None,
                 parent_id=None,
                 name=None,
                 adunits=None,
                 external_id=None,
                 external_name=None):
        self.id = id
        self.parent_id = parent_id
        self.name = name
        self.adunits = adunits
        self.external_id = external_id
        self.external_name = external_name


class Geography(ObjectModel):
    """
    Abstract Class for geography targets
    """
    def __init__(self,
                 id=None,
                 parent_id=None,
                 type=None,
                 name=None,
                 external_id=None,
                 external_name=None):
        self.id = id
        self.parent_id = parent_id
        self.type = type
        self.name = name
        self.external_id = external_id
        self.external_name = external_name


class Technology(ObjectModel):
    """
    Abstract Class for technology targets
    """
    def __init__(self,
                 id=None,
                 parent_id=None,
                 name=None,
                 external_id=None,
                 external_name=None,
                 type=None,
                 version=None):
        self.id = id
        self.parent_id = parent_id
        self.name = name
        self.external_id = external_id
        self.external_name = external_name
        self.type = type
        self.version = version


class DayPart(ObjectModel):
    """
    Abstract Class for day part targets
    """
    def __init__(self,
                 day_of_week,
                 start,
                 end):
        self.day_of_week = day_of_week
        self.start = start
        self.end = end


class UserDomain(ObjectModel):
    """
    Abstract Class for user domain targets
    """
    def __init__(self,
                 domain):
        self.domain = domain


class Custom(ObjectModel):
    """
    Abstract Class for custom targets
    """
    def __init__(self,
                 id=None,
                 id_key=None,
                 parent_id=None,
                 name=None,
                 type=None,
                 external_id=None,
                 external_name=None,
                 descriptive_name=None,
                 node_key=None):
        self.id = id
        self.id_key = id_key
        self.parent_id = parent_id
        self.name = name
        self.type = type
        self.external_id = external_id
        self.external_name = external_name
        self.descriptive_name = descriptive_name
        self.node_key = node_key


class VideoContent(ObjectModel):
    """
    Abstract Class for video content targets
    """
    def __init__(self,
                 id=None,
                 parent_id=None,
                 name=None,
                 external_id=None,
                 external_name=None):
        self.id = id
        self.parent_id = parent_id
        self.name = name
        self.external_id = external_id
        self.external_name = external_name


class VideoPosition(ObjectModel):
    """
    Abstract Class for video position targets
    """
    def __init__(self,
                 id=None,
                 parent_id=None,
                 name=None,
                 external_id=None,
                 external_name=None):
        self.id = id
        self.parent_id = parent_id
        self.name = name
        self.external_id = external_id
        self.external_name = external_name


class TargetingCriterion(object):
    """
    Class for creating boolean combinations of target models
    """

    OPERATOR = Enum(['OR', 'AND', 'NOT'])
    """
    Enum, list of allowed operators
    """

    def __init__(self, target_list, operator=None):
        """
        @param target_list: list(ObjectModel)|list(TargetingCriterion)|ObjectModel
        @param operator: self.OPERATOR
        """
        if isinstance(target_list, ObjectModel):
            # Cast to list in case a single ObjectModel is inputed
            target_list = [target_list]
            operator = self.OPERATOR.OR

        if not operator in self.OPERATOR:
            raise ParselmouthException("Invalid operator")
        if not isinstance(target_list, list):
            raise ParselmouthException("Invalid target list")

        self._data = {operator: target_list}

    def __str__(self):
        return "{class_name}({data})".format(
            class_name=self.__class__.__name__,
            data=str(self._data),
        )

    def __repr__(self):
        return str(self)

    @classmethod
    def _add_criterion(cls, first, second, operator):
        """
        Add two criterion with & or |. If there would be
        a nested AND: { AND: { present in new criterion,
        squash the result into AND: [].

        @param first: TargetingCriterion
        @param second: TargetingCriterion
        @param operator: TargetingCriterion.OPERATOR
        @return: TargetingCriterion
        """
        assert isinstance(first, TargetingCriterion)
        assert isinstance(second, TargetingCriterion)
        assert operator in cls.OPERATOR

        op1, targets1 = first.get_data()
        op2, targets2 = second.get_data()

        first_is_extendable = op1 == operator or len(targets1) <= 1
        second_is_extendable = op2 == operator or len(targets2) <= 1
        # Are the targets of the same python type?
        same_types = type(targets1[0]) == type(targets2[0])

        if op1 == cls.OPERATOR.NOT or op2 == cls.OPERATOR.NOT:
            # If either operator is a not, do not
            # attempt to simplify the response
            new_target_list = [first, second]
        elif first_is_extendable and second_is_extendable and same_types:
            # If both summands are in agreement with new operator,
            # and their types match, then concatenate their targets
            new_target_list = targets1 + targets2
        else:
            new_target_list = [first, second]

        return TargetingCriterion(
            new_target_list,
            operator,
        )

    def remove_target(self, target):
        """
        Remove the specified target from the current criterion.
        We assume that the criterion has a AND [] construction, and
        furthermore that the target lies within the top-level AND.

        @param criterion: TargetingCriterion
        @param target: TargetingCriterion

        @return: TargetingCriterion|None
        """
        assert isinstance(target, TargetingCriterion)

        if self == target:
            return None

        op1, targets1 = self.get_data()
        op2, targets2 = target.get_data()

        if not op1 == self.OPERATOR.AND:
            raise ParselmouthException(
                "TargetingCriterion has wrong structure. Top-level operator not an AND."
            )

        if not len(targets2) == 1 or isinstance(targets2[0], TargetingCriterion):
            raise ParselmouthException(
                "Target should be a simple TargetingCriterion with a single ObjectModel."
            )

        target_object_model = targets2[0]

        target_index = None
        if target_object_model in targets1:
            target_index = targets1.index(target_object_model)

        if target in targets1:
            target_index = targets1.index(target)

        if target_index is None:
            raise ParselmouthException(
                "Target not found in top-level structure."
            )

        targets1 = deepcopy(targets1)
        targets1.pop(target_index)

        # Handle the following case which creates an extraneous AND:
        # AND: [TargetingCriterion] ---> TargetingCriterion
        if len(targets1) == 1 and isinstance(targets1[0], TargetingCriterion):
            return targets1[0]

        # Default to OR operator for single target:
        # AND: [ObjectModel] ---> OR: [ObjectModel]
        if len(targets1) == 1:
            op1 = self.OPERATOR.OR

        return TargetingCriterion(
            targets1,
            op1,
        )

    def __and__(self, other):
        return self._add_criterion(
            self, other, self.OPERATOR.AND,
        )

    def __or__(self, other):
        return self._add_criterion(
            self, other, self.OPERATOR.OR,
        )

    def __invert__(self):
        return TargetingCriterion(
            [self],
            self.OPERATOR.NOT,
        )

    def __ne__(self, other):
        return not(self == other)

    def __eq__(self, other):
        if not isinstance(other, TargetingCriterion):
            return False
        return check_equal(self._data, other._data)

    def get_data(self):
        """
        @return: operation, target_list
        """
        return self._data.items()[0]

    def to_doc(self):
        """
        Serialize this object into a dictionary

        @return: dict
        """
        doc = {}
        if not self._data:
            return doc

        for operator, _values in self._data.items():
            doc[operator] = [d.to_doc() for d in _values]

        # Save metadata for use when re-cerealizing data
        doc['_metadata'] = {
            'cls': self.__class__.__name__,
        }
        return doc

    @classmethod
    def from_doc(cls, doc):
        """
        Convert a dictionary (from to_doc) back to its
        native ObjectModel type

        @param doc: dict
        """
        tmp_doc = deepcopy(doc)
        del tmp_doc['_metadata']
        operator, doc_list = tmp_doc.items()[0]

        target_list = []
        for _doc in doc_list:
            if _doc['_metadata']['cls'] == 'TargetingCriterion':
                target_list.append(TargetingCriterion.from_doc(_doc))
            else:
                target_list.append(ObjectModel.from_doc(_doc))

        return cls(target_list, operator)

    def flatten(self):
        """
        Return flattened list of all targeting
        objects in self.  By default all targeting objects are
        returned.  Only one of includes_only or excludes_only
        may be True.

        @param includes_only: bool, only give ObjectModels that
            do not live under a NOT operator
        @param excludes_only: bool, only give ObjectModels that
            live under a NOT operator
        @return: list(ObjectModel)
        """
        models = []
        _, targets = self.get_data()
        for t in targets:
            if isinstance(t, ObjectModel):
                models += [t]
            elif isinstance(t, TargetingCriterion):
                models += t.flatten()

        return models

    def _get_includes(self):
        """
        Recursive function to get all targeting objects
        that do not live under a NOT operator

        @return: list(ObjectModel)
        """
        models = []
        op, targets = self.get_data()
        if op == self.OPERATOR.NOT:
            return models

        for t in targets:
            if isinstance(t, ObjectModel):
                models += [t]
            elif isinstance(t, TargetingCriterion):
                if op != self.OPERATOR.NOT:
                    models += t._get_includes()

        return models

    def get_includes_and_excludes(self):
        """
        Get a list of all targeting objects that
        are included in the target (do not live under a NOT),
        and a list of excluded targets(live under a NOT).

        WARNING: This functionality only works for simple
        targeting criterion that do not have very complex
        boolean construction. This only works for targeting
        criterion that have a scheme like:
            (OR: [1,2,3]) & (NOT: OR: [4,5,6])
        In more complex targeting setups, you cannot guarantee that
        this function will output what you actually want.

        @return: list(ObjectModel), list(ObjectModel)
        """
        # Get set of all targeting objects
        all_targets = set(self.flatten())
        # Get set of all included objects
        includes = set(self._get_includes())
        # Get the items from all_targets which are not in includes
        excludes = all_targets - includes
        return list(includes), list(excludes)


class TargetingData(ObjectModel):
    """
    Abstract Class for line item targeting information
    Targeting indicates which kinds of users get served
    and ad impression or which part of a hosts site to
    serve an ad to
    """

    def __init__(self,
                 inventory=None,
                 geography=None,
                 day_part=None,
                 user_domain=None,
                 technology=None,
                 video_content=None,
                 video_position=None,
                 custom=None):
        self.inventory = inventory
        self.geography = geography
        self.day_part = day_part
        self.user_domain = user_domain
        self.technology = technology
        self.video_content = video_content
        self.video_position = video_position
        self.custom = custom

    def to_doc(self):
        """
        Serialize this object into a dictionary

        @return: dict
        """
        doc = {}
        for _key, _value in vars(self).items():
            if isinstance(_value, TargetingCriterion):
                doc[_key] = _value.to_doc()
            else:
                doc[_key] = _value

        # Save metadata for use when re-cerealizing data
        doc['_metadata'] = {
            'cls': self.__class__.__name__,
        }
        return doc

    @classmethod
    def from_doc(cls, doc):
        """
        Convert a dictionary (from to_doc) back to its
        native ObjectModel type

        @param doc: dict
        """
        params = {}
        for _key, _value in doc.items():
            if _key == '_metadata':
                continue
            elif _value:
                params[_key] = TargetingCriterion.from_doc(_value)

        return cls(**params)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - Ad Server Object Models

Base Interfaces for ad server items
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
from abc import ABCMeta
from pprint import pformat
import logging

# Global Variable Definitions
_class_name_map = None
"""
dict, global variable used to contain mappings of the string name for an
      object to its class
"""


def _initialize_class_name_map():
    """
    Create a dictionary mapping each ObjectModel
    child to its name.  This is used for cerealizing
    dictionary objects back to ObjectModels
    """
    global _class_name_map

    from parselmouth.delivery import Cost
    from parselmouth.delivery import Goal
    from parselmouth.delivery import Stats
    from parselmouth.delivery import DeliveryMeta
    from parselmouth.delivery import Creative
    from parselmouth.delivery import LineItem
    from parselmouth.delivery import Campaign
    from parselmouth.targeting import TargetingData
    from parselmouth.targeting import AdUnit
    from parselmouth.targeting import Placement
    from parselmouth.targeting import Geography
    from parselmouth.targeting import Technology
    from parselmouth.targeting import DayPart
    from parselmouth.targeting import UserDomain
    from parselmouth.targeting import Custom
    from parselmouth.targeting import VideoContent
    from parselmouth.targeting import VideoPosition
    from parselmouth.targeting import TargetingCriterion

    _class_name_map = {
        'Cost': Cost,
        'Goal': Goal,
        'Stats': Stats,
        'DeliveryMeta': DeliveryMeta,
        'Creative': Creative,
        'LineItem': LineItem,
        'Campaign': Campaign,
        'TargetingData': TargetingData,
        'AdUnit': AdUnit,
        'Placement': Placement,
        'Geography': Geography,
        'Technology': Technology,
        'DayPart': DayPart,
        'UserDomain': UserDomain,
        'Custom': Custom,
        'VideoContent': VideoContent,
        'VideoPosition': VideoPosition,
        'TargetingCriterion': TargetingCriterion,
    }


class ObjectModel(object):
    """
    Abstract Base for all parselmouth object models
    """

    __metaclass__ = ABCMeta
    """
    Ensure that this abstract class is extended
    """

    ignored_comparable_keys = [
        'last_modified',
        'last_modified_by',
        'preview_url',
    ]
    """
    list(str), list of keys that will be ignored when comparing objects

    NOTES:
        * Unfortunately, google DFP appears to encode a hash into the
        preview url response of a creative. This is likely part of some
        measure to enforce a TTL on an asset or prevent it from being
        scraped continuously. This means that we can't really check
        that the urls are the same :(
            One thing we could do to get around this is to store the
            base64 encoding of the image in the object we test against
            and download the asset from the preview url given by DFP.
            Then all we need to do is compare the base64 encoded assets.

    """

    def __str__(self):
        """
        Human readable representation of this object
        @return: str
        """
        pretty = False
        return "{class_name}({vars})".format(
            class_name=self.__class__.__name__,
            vars=pformat(vars(self)) if pretty else str(vars(self))
        )

    def __repr__(self):
        return str(self)

    def __ne__(self, other):
        return not(self == other)

    def __getitem__(self, key):
        return vars(self)[key]

    def __eq__(self, other):
        if isinstance(other, ObjectModel):
            self_dict = vars(self)
            other_dict = vars(other)
            for _key, _value in self_dict.items():

                # Skip comparison of keys like "last_modified" that
                # don't really communicate anything useful
                if _key in self.ignored_comparable_keys:
                    continue

                if _key not in other_dict:
                    logging.warning(
                        "{0} key is not in the other object".format(_key)
                    )
                    return False
                elif other_dict.get(_key) != _value:
                    logging.debug(
                        "Other value ({0}) does not equal this value ({1})".format(
                            other_dict.get(_key),
                            _value
                        )
                    )
                    return False
            return True
        return NotImplemented

    def to_doc(self):
        """
        Serialize this object into a dictionary

        @return: dict
        """
        doc = {}
        for _key, _value in vars(self).items():
            if isinstance(_value, ObjectModel):
                doc[_key] = _value.to_doc()
            else:
                doc[_key] = _value

        # Save metadata for use when re-cerealizing data
        doc['_metadata'] = {
            'cls': self.__class__.__name__,
        }
        return doc

    @classmethod
    def from_doc(cls, doc):
        """
        Convert a dictionary (from to_doc) back to its
        native ObjectModel type

        @param doc: dict
        """
        if not _class_name_map:
            _initialize_class_name_map()

        params = {}
        for _key, _value in doc.items():
            if _key == '_metadata':
                continue
            elif isinstance(_value, dict):
                cls_name = _value['_metadata']['cls']
                params[_key] = _class_name_map[cls_name].from_doc(_value)
            else:
                params[_key] = _value

        # In the case when ObjectModel.from_doc is called
        # outside of the child class, read the class name
        # from metadata and call that classes init function
        _doc_cls_name = doc['_metadata']['cls']
        if _class_name_map.get(_doc_cls_name):
            _doc_cls = _class_name_map[_doc_cls_name]
        else:
            _doc_cls = cls

        return _doc_cls(**params)

    def __hash__(self):
        return hash(frozenset(vars(self)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Parselmouth utilities - Enum
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class Enum(set):
    """
    Usage:
    MyEnum = Enum([
    'foo',
    'bar',
    ])

    Then you can do:
    MyEnum.foo and MyEnum.bar
    """
    def __getattr__(self, name):
        if name in self:
            return name
        raise KeyError('invalid name: %s' % name)

    def __getitem__(self, name):
        return self.__getattr__(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Google DFP Utils
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Third Party Library Imports
from suds.sudsobject import asdict


def _underscore_key(string):
    """
    Replace periods with underscores

    @return string, str
    @return key, str
    """
    return string.replace('.', '_')


def to_string(item, convert_bool=True):
    """
    Handles unicode strings with non-ascii characters
    """
    str_val = unicode(item).encode('utf-8', 'replace')

    if convert_bool:
        if str_val == u'True' or str_val == u'true':
            return True
        elif str_val == u'False' or str_val == u'false':
            return False

    return str_val


def recursive_asdict(obj):
    """
    Convert Suds object into a dict so it can be serialized.
    Taken (mostly) from: http://stackoverflow.com/a/15678861

    @param obj, sudsobject: the suds object to be converted to dict
    @return dict: the object converted to a dict
    """
    # Don't convert a dict
    if isinstance(obj, dict):
        return obj

    out = {}
    for key, val in asdict(obj).iteritems():
        if hasattr(val, '__keylist__'):
            out[_underscore_key(key)] = recursive_asdict(val)
        elif isinstance(val, list):
            out[_underscore_key(key)] = []
            for item in val:
                if hasattr(item, '__keylist__'):
                    out[_underscore_key(key)].append(recursive_asdict(item))
                else:
                    out[_underscore_key(key)].append(to_string(item))
        else:
            out[_underscore_key(key)] = to_string(val)
    return out


def format_pql_response(raw_list):
    """
    Convert a list of lists to where the first
    row is a list of column names, to a list
    of dictionaries with keys associated to each
    column name and values the associated values.

    @param raw_list: list
    @return: list(dict)
    """
    column_keys = raw_list.pop(0)
    data_list = []
    for result in raw_list:
        data = {}
        for col_name, val in zip(column_keys, result):
            # NOTE: DFP returns numbers as integers. We refer to integers
            # as strings throughout the app and therefore want to be
            # consistent here.
            data[col_name] = str(val)
        data_list.append(data)
    return data_list


def format_report_list(raw_list):
    """
    Convert a list of lists to where the first
    row is a list of column names, to a list
    of dictionaries with keys associated to each
    column name and values the associated values.

    @param raw_list: list
    @return: list(dict)
    """
    column_keys = raw_list.pop(0)
    data_list = []
    for result in raw_list:
        data = {}
        for col_name, val in zip(column_keys, result):
            # NOTE: DFP returns numbers as integers. We refer to integers
            # as strings throughout the app and therefore want to be
            # consistent here.
            val = val.encode('utf-8')
            data[col_name] = str(val)
        data_list.append(data)
    return data_list


def sanitize_report_response(response):
    """
    The dfp report api returns reponses with
    types pre-pended to column names. This function
    deletes these to make for a cleaner response.

    @param response: dict
    @return: dict
    """
    clean_response = {}
    for _key, _val in response.items():
        clean_key = _key.split('Dimension.')[-1]
        clean_key = clean_key.split('Column.')[-1]
        clean_response[clean_key] = _val

    return clean_response



"""
Set of constants for pulling data from DFP
"""

# Chartbeat Imports
from parselmouth.constants import AdProviderSellTypes
from parselmouth.constants import ParselmouthReportMetrics
from parselmouth.utils.enum import Enum


DFP_API_VERSION = 'v201508'
"""
str, version of the DFP API to use
https://developers.google.com/doubleclick-publishers/
"""

DFP_QUERY_DEFAULTS = {
    'order': 'ID',
    'limit': 1,
    'offset': 0
}
"""
dict, default query params for the DFP API
"""

SELL_TYPE_MAP = {
    'SPONSORSHIP': AdProviderSellTypes.sponsorship,
    'STANDARD': AdProviderSellTypes.standard,
    'NETWORK': AdProviderSellTypes.network,
    'HOUSE': AdProviderSellTypes.house,
    'PRICE_PRIORITY': AdProviderSellTypes.price_priority,
}
"""
dict, map dfp sell type names -> AD_PROVIDER_SELL_TYPES
"""

DFP_REPORT_METRIC_MAP = {
    ParselmouthReportMetrics.ad_impressions: "AD_SERVER_IMPRESSIONS",
    ParselmouthReportMetrics.ad_viewable_impressions: "AD_SERVER_ACTIVE_VIEW_VIEWABLE_IMPRESSIONS",
    ParselmouthReportMetrics.line_item_id: "LINE_ITEM_ID",
    ParselmouthReportMetrics.line_item_name: "LINE_ITEM_NAME",
    ParselmouthReportMetrics.delivery_percentage: "AD_SERVER_DELIVERY_INDICATOR",
}
"""
dict, map to column/dimension names in the dfp report api
"""

DFP_CUSTOM_TARGETING_KEY_TYPES = Enum([
    'PREDEFINED',
    'FREEFORM'
])
"""
Enum, list of the available custom targeting key types
"""

DFP_VALUE_MATCH_TYPES = Enum([
    'EXACT',
    'BROAD',
    'PREFIX',
    'BROAD_PREFIX',
    'SUFFIX',
    'CONTAINS',
    'UNKNOWN'
])
"""
Enum, list of the available custom targeting value match types
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth Abstract Configuration
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
from abc import ABCMeta
from abc import abstractmethod


class AbstractConfig(object):
    """
    Abstract Configuration to all Ad Services

    This class outlines functions required to be implemented by ad
    service configuration classes. This class should serve as a way
    to authenticate to a given ad provider service.
    If that functionality is not provided by a specific ad
    service configuration then it needs to raise `NotImplementedError`.

    Right now most of this functionality follows the DFP implementation.
    It's likely that many of these functions will not be used by other
    ad services.
    """
    __metaclass__ = ABCMeta

    def __repr__(self):
        """
        Human readable representation of this object

        @return: str
        """
        return (
            "{class_name}("
            ")"
        ).format(
            class_name=self.__class__.__name__,
        )

    @abstractmethod
    def validate_credentials(self):
        pass

    @abstractmethod
    def get_credentials_arguments(self):
        pass

    @abstractmethod
    def load_config_from_file(self, config_path):
        pass

    @abstractmethod
    def load_config_from_external_source(self, *args, **kwargs):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - Contants

Constants related to Parselmouth and the providers that use it
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Parselmouth Imports
from parselmouth.utils.enum import Enum


ParselmouthProviders = Enum([
    'google_dfp_premium',
    'google_dfp_small_business',
])
"""
Enum, List of providers that Parselmouth interfaces with
"""

ParselmouthTargetTypes = Enum([
    'adunit',
    'geography',
    'demographics',
    'ad_position',
    'custom',
])
"""
Enum, list of target types
"""

TechnologyTargetTypes = Enum([
    'bandwidth_group',
    'browser',
    'browser_language',
    'device_capability',
    'device_category',
    'device_manufacturer',
    'mobile_carrier',
    'mobile_device',
    'mobile_device_submodel',
    'operating_system',
    'operating_system_version',
])
"""
Enum, list of valid technology types for targeting
"""


ParselmouthReportMetrics = Enum([
    'ad_impressions',
    'ad_viewable_impressions',
    'line_item_id',
    'line_item_name',
    'delivery_percentage',
])
"""
Enum, metrics which can be pulled from ad provider reports
"""


AdProviderSellTypes = Enum([
    'sponsorship',
    'standard',
    'network',
    'house',
    'price_priority',
])
"""
Enum, sales categories for campaigns
"""


MAX_REQUEST_ATTEMPTS = 3



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - Main Package Entry Point

This package serves as the base interface between Chartbeat worker and
services and third-party ad servers.
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
import logging

# Parselmouth Imports
from parselmouth.constants import MAX_REQUEST_ATTEMPTS
from parselmouth.constants import ParselmouthProviders
from parselmouth.constants import ParselmouthReportMetrics
from parselmouth.exceptions import ParselmouthException
from parselmouth.exceptions import ParselmouthNetworkError
from parselmouth.tree_builder import TreeBuilder
from parselmouth.utils.timeout import Timeout
from parselmouth.config import ParselmouthConfig

# Parselmouth Imports - Adapter Imports
from parselmouth.adapters.dfp.interface import DFPInterface


class Parselmouth(object):
    """
    Base Interface Class
    """

    ProviderInterfaceMap = {
        ParselmouthProviders.google_dfp_premium: DFPInterface,
        ParselmouthProviders.google_dfp_small_business: DFPInterface,
    }
    """
    dict, mapping between ad service providers and their implementations
    NOTES:
        * This should probably be hosted elsewhere...
        * This doesn't serve much purpose right now since all we support
            is DFP...
    """

    def __init__(self,
                 config=None,
                 provider_name=None,
                 network_timeout=60 * 10,
                 **kwargs):
        """
        Constructor

        Authenticates a provider client for the given domain.

        @param config: parselmouth.config.ParselmouthConfig
        @param provider_name: any(parselmouth.constants.ParseltoungProvider)
        @param network_timeout: int, number seconds before timing out a request
            the given ad provider service
        """
        self._network_timeout = network_timeout
        self.provider_config = config
        # Load the provider configuration
        if self.provider_config and not isinstance(self.provider_config, ParselmouthConfig):
            raise ParselmouthException(
                "Invalid config.  Config should be of type ParselmouthConfig",
            )
        elif not self.provider_config:
            self.provider_config = ParselmouthConfig(provider_name, **kwargs)

        # Get interface for given provider
        self.provider_name = self.provider_config.provider_name
        provider_interface_class = self.get_ad_service_interface_for_provider(
            self.provider_name
        )

        self.provider = provider_interface_class(
            **self.provider_config.get_credentials_arguments()
        )
        self.tree_builder = TreeBuilder(
            self.provider_name,
            self.provider,
        )

        # Attempt to access the network to check proper configuration
        try:
            self.get_network_timezone()
        except Exception as e:
            raise ParselmouthException("Provider not configured correctly. {}".format(str(e)))

    def __str__(self):
        """
        Human readable representation of this object

        @return: str
        """

        return (
            "{class_name}("
                "provider={provider},"
                "provider_config={provider_config}"
            ")"
        ).format(
            class_name=self.__class__.__name__,
            provider=self.provider,
            provider_config=self.provider_config,
        )

    @classmethod
    def get_ad_service_interface_for_provider(cls, provider_name):
        """
        Staticmethod, returns the interface class for a given ad service
        provider name

        @param provider_name: str, one of enum
            parselmouth.constants.ParselmouthProviders
        @return: descendant of
            parselmouth.adapters.abstract_interface.AbstractInterface
        """
        client_interface = cls.ProviderInterfaceMap.get(provider_name)
        if not client_interface:
            raise ValueError(
                "There is no interface defined for provider: %s" %
                provider_name
            )
        return client_interface

    @classmethod
    def get_ad_service_config_interface_for_provider(cls, provider_name):
        """
        Staticmethod, returns the config interface for a given ad
        service provider name

        @param provider_name: str, one of enum
            parselmouth.constants.ParselmouthProviders
        @return: child(parselmouth.config.ParselmouthConfig)
        """
        config_interface = cls.ProviderConfigInterfaceMap.get(provider_name)
        if not config_interface:
            raise ValueError(
                "There is no config interface defined for provider: %s" %
                provider_name
            )
        return config_interface

    def get_network_timezone(self):
        """
        Get the DFP network timezone for
        the host this client is connected to

        @return: pytz.timezone
        """
        with Timeout(self._network_timeout):
            return self.provider.get_network_timezone()

    def get_advertisers(self):
        """
        Queries dfp for all advertisers within their account

        @return: list(dict), returns a list of company dictionaries
        """
        with Timeout(self._network_timeout):
            return self.provider.get_advertisers()

    def get_campaign(self, campaign_id, include_line_items=False):
        """
        Return a campaign object given an id

        @param campaign_id: str, id of the campaign to return
        @param include_line_items: bool, include line item data as well
        @return: parselmouth.delivery.Campaign
        """
        with Timeout(self._network_timeout):
            campaign = self.provider.get_campaign(campaign_id)

        if include_line_items:
            campaign.line_items = self.get_campaign_line_items(campaign)

        return campaign

    def get_campaigns(self, **kwargs):
        """
        Get campaigns on optional filters

        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: L{parselmouth.delivery.Campaign}
        """
        with Timeout(self._network_timeout):
            return self.provider.get_campaigns(**kwargs)

    def get_line_item(self, line_item_id):
        """
        Return a line item object given an id

        @param line_item_id: str, id of the LineItem to return
        @return: parselmouth.delivery.LineItem
        """
        attempt = 1
        response = None
        while not response and attempt <= MAX_REQUEST_ATTEMPTS:
            try:
                with Timeout(self._network_timeout):
                    response = self.provider.get_line_item(line_item_id)
            except ParselmouthNetworkError:
                logging.exception("Got network error on attempt %s" % attempt)
                attempt += 1

        if not response:
            raise ParselmouthException(
                'Could not fetch a line item in {0} attempts'.format(
                    MAX_REQUEST_ATTEMPTS
                )
            )

        return response

    def get_line_items(self, **kwargs):
        """
        Get line items on optional filters

        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: L{parselmouth.delivery.LineItem}
        """
        with Timeout(self._network_timeout):
            return self.provider.get_line_items(**kwargs)

    def get_campaign_line_items(self, campaign):
        """
        Get line items on optional filters

        @param campaign: Campaign|str,
        @return: L{parselmouth.delivery.LineItem}
        """
        with Timeout(self._network_timeout):
            return self.provider.get_campaign_line_items(campaign)

    def get_line_item_available_inventory(self,
                                          line_item,
                                          use_start=False,
                                          preserve_id=False):
        """
        Get number of impressions available for line item.
        NOTE: The following fields are required within a line item to
        perform an availability check:
            type, cost_type, end, targeting.inventory

        @param line_item: LineItem
        @param use_start: bool, if False, checks availability from right
            now
        @param preserve_id: bool, communicate the id of the line item
            being forecasted to DFP. This is used in the case where we
            want to get a forecast on a line item in flight. This way, a
            domain doesn't need to have enough spare inventory to
            accommodate the two line items simultaneously. NOTE: If this
            is true then use_start is necessarily true.
        @return: int|None, number of available impressions
        """
        with Timeout(self._network_timeout):
            return self.provider.get_line_item_available_inventory(
                line_item, use_start, preserve_id
            )

    def get_creative(self, creative_id):
        """
        Return a creative object given an id

        @param creative_id: str, id of the campaign to return
        @return: parselmouth.delivery.Creative
        """
        return self.provider.get_creative(creative_id)

    def get_creatives(self, **kwargs):
        """
        Get creatives on optional filters

        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: L{parselmouth.delivery.Creative}
        """
        with Timeout(self._network_timeout):
            return self.provider.get_creatives(**kwargs)

    def get_line_item_creatives(self, line_item):
        """
        Return the creatives associated with a given line item

        @param line_item: str|int|parselmouth.delivery.LineItem,
            either the id of the lineitem or an object with the id
        @return: list(parselmouth.delivery.Creative)
        """
        with Timeout(self._network_timeout):
            return self.provider.get_line_item_creatives(line_item)

    def get_line_item_report(self,
                             start,
                             end,
                             columns=[ParselmouthReportMetrics.ad_impressions]):
        """
        Get the number of impressions served by DFP
        for all line items between the two datetimes.

        Note: These datetimes are assumed to be in the timezone
        of the host and will be aligned to day

        Note: This API has a strange behavior where the date
        inputs only allow for a calendar day.
        This means start=1/1, end=1/1 will query against
        the entire day of 1/1

        @param start: datetime,
        @param end: datetime,
        @param columns: list(str), list of columns to include
        @return: list(dict)
        """
        with Timeout(self._network_timeout):
            return self.provider.get_line_item_report(
                start, end, columns,
            )

    def get_custom_target_by_name(self, name, parent_name):
        """
        Get a custom target by its name

        @param name: name of custom target
        @param parent_name: name of parent target
        @return: Custom|None
        """

        with Timeout(self._network_timeout):
            targets = self.provider.get_custom_targets(
                key_name=parent_name, value_name=name,
            )

        if targets and len(targets) > 1:
            raise ParselmouthException('Given name is not unique')
        elif targets:
            return targets[0]
        else:
            return None

    def construct_tree(self, target_type):
        """
        Get all data of type target_type from ad provider,
        and build a tree

        @param taget_type: parselmouthTargetTypes
        @return: NodeTree
        """
        return self.tree_builder.construct_tree(target_type)

    def update_line_item(self, line_item):
        """
        Update a Line Item for a provider

        @param line_item: parselmouth.delivery.LineItem
        """
        with Timeout(self._network_timeout):
            self.update_line_items([line_item])

    def update_line_items(self, line_items):
        """
        Update multiple Line Item for a provider

        @param line_items: L{parselmouth.delivery.LineItem}
        """
        with Timeout(self._network_timeout):
            self.provider.update_line_items(line_items)

    def create_custom_target(self, key, value):
        """
        Add a custom target to this domain's ad service provider

        @param key: parselmouth.targeting.Custom
        @param value: parselmouth.targeting.Custom
        @return: list(parselmouth.targeting.Custom)
        """
        with Timeout(self._network_timeout):
            return self.provider.create_custom_target(key, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Parselmouth utilities
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
import signal

# Local Package Imports
from parselmouth.exceptions import ParselmouthTimeout


class Timeout(object):
    """
    Timeout context manager class for wrapping function calls that may
    not complete.

    This code is adapted from the gist available here:
        https://gist.github.com/glenfant/7501911

    Warnings:
    - This does not work with Windows that does not handle the signals we need.
    - This is not thead safe since the signal will get caught by a random thread.
    - Tested on MacOSX with Python 2.6, 2.7 and 3.3 (may or not work eslsewhere)
    """

    def __init__(self, timeout, swallow_exception=False):
        """
        @param timeout: int, seconds enabled for processing the block
            under our context manager
        @param swallow_exception: bool, do not raise the exception on
            timeout
        """
        self._timeout = timeout
        self._swallow_exception = swallow_exception

    def __enter__(self):
        self.setup_timeout()

    def __exit__(self, exc_type, value, traceback):
        # Always cancel immediately, since we're done
        try:
            self.cancel_timeout()
        except ParselmouthTimeout:
            # Weird case: we're done with the with body, but now the
            # alarm is fired.  We may safely ignore this situation and
            # consider the body done.
            pass

        # __exit__ may return True to supress further exception
        # handling.  We don't want to suppress any exceptions here,
        # since all errors should just pass through, ParselmouthTimeout
        # being handled normally to the invoking context.
        if exc_type is ParselmouthTimeout and self._swallow_exception:
            return True
        return False

    def handle_timeout(self, signum, frame):
        raise ParselmouthTimeout(
            '%s exceeded maximum timeout value (%d seconds).' % (
                frame.f_code.co_name,
                self._timeout
            )
        )

    def setup_timeout(self):
        """
        Sets up an alarm signal and a signal handler that raises a
        ParselmouthTimeout after the timeout amount (expressed in
        seconds).
        """
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self._timeout)

    def cancel_timeout(self):
        """
        Removes the death penalty alarm and puts back the system into
        default signal handling.
        """
        signal.alarm(0)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)



from setuptools import setup, find_packages

setup(
    name='parselmouth',
    packages=find_packages(),
    version='1.0.2',
    description=(
        'An object oriented interface for ad provider services in python'
    ),
    author='Paul Kiernan, Justin Mazur',
    author_email='paulkiernan1@gmail.com, justindmazur@gmail.com',
    url='https://github.com/chartbeat-labs/parselmouth',
    download_url='https://github.com/chartbeat-labs/parselmouth/tarball/1.0.2',
    keywords = ['googleads', 'dfp', 'ad server', 'parselmouth', 'ad provider'],
    classifiers = [],
    install_requires=[
        'googleads==3.8.0',
        'pytz==2015.7',
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Google DFP Utils
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
import logging
from datetime import datetime

# Third Party Library Imports
import pytz

# Parselmouth Imports
from parselmouth.delivery import Campaign
from parselmouth.delivery import Creative
from parselmouth.delivery import Cost
from parselmouth.delivery import DeliveryMeta
from parselmouth.delivery import Goal
from parselmouth.delivery import LineItem
from parselmouth.delivery import Stats
from parselmouth.adapters.dfp.constants import SELL_TYPE_MAP
from parselmouth.adapters.dfp.targeting_utils import transform_targeting_data_from_dfp
from parselmouth.adapters.dfp.targeting_utils import transform_targeting_data_to_dfp


def dfp_date_to_datetime(dfp_date_dict):
    """
    Convert a DFP dictionary representation of a date into a native
    python datetime object

    @param dfp_date_dict: dict, dictionary representation of a date
    @return: datetime.datetime, native datetime object of the input date
    """

    # Extract
    year = dfp_date_dict['date']['year']
    month = dfp_date_dict['date']['month']
    day = dfp_date_dict['date']['day']
    hour = dfp_date_dict['hour']
    minute = dfp_date_dict['minute']
    second = dfp_date_dict['second']
    timezone_id = dfp_date_dict['timeZoneID']

    # Validate and cast
    year = int(year)
    month = int(month)
    day = int(day)
    hour = int(hour)
    minute = int(minute)
    second = int(second)
    tzinfo = pytz.timezone(timezone_id)

    return datetime(
        year=year,
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
        tzinfo=tzinfo
    )


def datetime_to_dfp_date(dt):
    """
    Convert datetime to dfp style date. If there is no tz
    UTC is assumed

    @param dt: datetime
    @return: dict
    """
    if not dt.tzinfo:
        dt = dt.replace(tzinfo=pytz.utc)

    dfp_date = {
        'timeZoneID': str(dt.tzinfo),
        'date': {
            'year': dt.year,
            'day': dt.day,
            'month': dt.month,
        },
        'second': dt.second,
        'minute': dt.minute,
        'hour': dt.hour,
    }
    return dfp_date


def _get_dfp_dates_from_object(dfp_dict):
    """
    Helper function for extracting dfp datetimes from DFP responses

    @param dfp_dict: dictionary with one of the keys `startDateTime`,
        `endDateTime`, `lastModifiedDateTime`
    @return: (start_date, end_date, last_modified_date)
    """

    start_datetime_dict = dfp_dict.get('startDateTime')
    if start_datetime_dict:
        start_datetime = dfp_date_to_datetime(start_datetime_dict)
    else:
        start_datetime = None

    end_datetime_dict = dfp_dict.get('endDateTime')
    if end_datetime_dict:
        end_datetime = dfp_date_to_datetime(end_datetime_dict)
    else:
        end_datetime = None

    last_modified_datetime = \
        dfp_date_to_datetime(dfp_dict.get('lastModifiedDateTime'))

    return (start_datetime, end_datetime, last_modified_datetime)


def transform_campaign_from_dfp(order):
    """
    Convert dictionary-representation for a SUDS order object into a
    Parselmouth representation of an order
    @param order: dict, dictionary-representation of a DFP SUDS response
        for a single order
    @return: parselmouth.delivery.Campaign
    """
    # Sanitize and cast inputs
    start_datetime, end_datetime, last_modified_datetime = \
        _get_dfp_dates_from_object(order)

    stats = Stats(
        impressions=order.get('totalImpressionsDelivered'),
        clicks=order.get('totalClicksDelivered')
    )
    total_budget = Cost(
        budget_micro_amount=order['totalBudget']['microAmount'],
        budget_currency_code=order['totalBudget']['currencyCode']
    )

    return Campaign(
        advertiser_id=order['advertiserId'],
        agency_id=order.get('agencyId'),
        creator_id=order['creatorId'],
        currency_code=order['currencyCode'],
        end=end_datetime,
        external_campaign_id=order['externalOrderId'],
        id=order['id'],
        last_modified=last_modified_datetime,
        last_modified_by=order['lastModifiedByApp'],
        name=order['name'],
        start=start_datetime,
        status=order['status'],
        stats=stats,
        total_budget=total_budget,
    )


def transform_campaign_to_dfp(campaign):
    """
    Transform a Campaign delivery object into
    a dictionary that will be accepted by dfp,

    @param campaign: Campaign
    @return: dict
    """
    return {
        'id': campaign.id,
        'name': campaign.name,
        'advertiserId': campaign.advertiser_id,
        'creatorId': campaign.creator_id,
        'status': campaign.status,
        'currencyCode': campaign.currency_code,
        'startDateTime': datetime_to_dfp_date(campaign.start),
        'endDateTime': datetime_to_dfp_date(campaign.end),
        'lastModifiedDateTime': datetime_to_dfp_date(campaign.last_modified),
        'lastModifiedByApp': campaign.last_modified_by,
        'totalBudget': {
            'microAmount': campaign.total_budget.budget_micro_amount,
            'currencyCode': campaign.total_budget.budget_currency_code,
        },
        'totalClicksDelivered': campaign.stats.clicks,
        'totalImpressionsDelivered': campaign.stats.impressions,
        'externalOrderId': campaign.external_campaign_id,
    }


def transform_line_item_from_dfp(line_item):
    """
    Convert dictionary-representation for a SUDS line item object into a
    Parselmouth representation of a Line Item

    @param order: dict, dictionary-representation of a DFP SUDS response
        for a single line item
    @return: parselmouth.delivery.LineItem
    """
    # Sanitize and cast inputs
    start_datetime, end_datetime, last_modified_datetime = \
        _get_dfp_dates_from_object(line_item)

    _delivery = line_item.get('deliveryIndicator', {})
    _stats = line_item.get('stats', {})
    stats = Stats(
        impressions=_stats.get('impressionsDelivered', 0),
        clicks=_stats.get('clicksDelivered', 0),
        video_starts=_stats.get('videoStartsDelivered', 0),
        video_completions=_stats.get('videoCompletionsDelivered', 0)
    )
    delivery_meta = DeliveryMeta(
        stats=stats,
        delivery_rate_type=line_item.get('deliveryRateType'),
        actual_delivery_percent=_delivery.get('actualDeliveryPercentage', 0),
        expected_delivery_percent=_delivery.get('expectedDeliveryPercentage', 0)
    )

    budget = Cost(
        budget_micro_amount=line_item['budget']['microAmount'],
        budget_currency_code=line_item['budget']['currencyCode']
    )
    cost_per_unit = Cost(
        budget_micro_amount=line_item['costPerUnit']['microAmount'],
        budget_currency_code=line_item['costPerUnit']['currencyCode']
    )
    value_cost_per_unit = Cost(
        budget_micro_amount=line_item['valueCostPerUnit']['microAmount'],
        budget_currency_code=line_item['valueCostPerUnit']['currencyCode']
    )
    primary_goal = Goal(
        goal_type=line_item['primaryGoal']['goalType'],
        unit_type=line_item['primaryGoal']['unitType'],
        units=int(line_item['primaryGoal']['units'])
    )

    targeting = transform_targeting_data_from_dfp(
        line_item['targeting'],
    )

    return LineItem(
        budget=budget,
        cost_per_unit=cost_per_unit,
        cost_type=line_item['costType'],
        delivery=delivery_meta,
        end=end_datetime,
        id=line_item['id'],
        last_modified=last_modified_datetime,
        last_modified_by=line_item['lastModifiedByApp'],
        type=SELL_TYPE_MAP.get(line_item['lineItemType']),
        name=line_item['name'],
        campaign_id=line_item['orderId'],
        campaign_name=line_item['orderName'],
        primary_goal=primary_goal,
        start=start_datetime,
        status=line_item['status'],
        targeting=targeting,
        value_cost_per_unit=value_cost_per_unit,
        creative_placeholder=line_item.get('creativePlaceholders'),
        target_platform=line_item.get('targetPlatform'),
    )


def _sell_type_reverse_lookup(line_item):
    """
    Get the dfp name for the given sell type

    @param line_item: LineItem
    @return: str
    """
    for dfp_type, native_type in SELL_TYPE_MAP.items():
        if line_item.type == native_type:
            return dfp_type

    return (line_item.type or '').upper()


def transform_line_item_to_dfp(line_item):
    """
    Transform a LineItem delivery object into
    a dictionary that will be accepted by dfp,

    @param line_item: LineItem
    @return: dict
    """
    return {
        'id': line_item.id,
        'name': line_item.name,
        'orderId': line_item.campaign_id,
        'orderName': line_item.campaign_name,
        'lineItemType': _sell_type_reverse_lookup(line_item),
        'costType': line_item.cost_type,
        'startDateTime': datetime_to_dfp_date(line_item.start),
        'endDateTime': datetime_to_dfp_date(line_item.end),
        'lastModifiedDateTime': datetime_to_dfp_date(line_item.last_modified),
        'lastModifiedByApp': line_item.last_modified_by,
        'budget': {
            'microAmount': "%.0f" % line_item.budget.budget_micro_amount,
            'currencyCode': line_item.budget.budget_currency_code,
        },
        'valueCostPerUnit': {
            'microAmount': "%.0f" % line_item.value_cost_per_unit.budget_micro_amount,
            'currencyCode': line_item.value_cost_per_unit.budget_currency_code,
        },
        'costPerUnit': {
            'microAmount': "%.0f" % line_item.cost_per_unit.budget_micro_amount,
            'currencyCode': line_item.cost_per_unit.budget_currency_code,
        },
        'primaryGoal': {
            'goalType': line_item.primary_goal.goal_type,
            'unitType': line_item.primary_goal.unit_type,
            'units': "%.0f" % line_item.primary_goal.units,
        },
        'creativePlaceholders': line_item.creative_placeholder,
        'stats': {
            'impressionsDelivered': line_item.delivery.stats.impressions,
            'clicksDelivered': line_item.delivery.stats.clicks,
            'videoStartsDelivered': line_item.delivery.stats.video_starts,
            'videoCompletionsDelivered': line_item.delivery.stats.video_completions,
        },
        'deliveryRateType': line_item.delivery.delivery_rate_type,
        'targetPlatform': line_item.target_platform,
        'targeting': transform_targeting_data_to_dfp(line_item.targeting),
        'status': line_item.status,
    }


def transform_forecast_line_item_to_dfp(line_item,
                                        use_start=False,
                                        preserve_id=False):
    """
    Transform a Line Item delivery object into
    a dictionary that will be accepted by dfp
    availability check

    @param line_item: LineItem
    @param use_start: bool
    @param preserve_id: bool, communicate the id of the line item being
        forecasted to DFP. This is used in the case where we want to get
        a forecast on a line item in flight. This way, a domain doesn't
        need to have enough spare inventory to accommodate the two line
        items simultaneously. NOTE: If this is true then use_start is
        necessarily true.
    @return: dict
    """
    assert line_item.type
    assert line_item.cost_type
    assert line_item.end
    # Inventory targeting is required for availability checks
    assert line_item.targeting.inventory

    dfp_doc = {
        'lineItemType': _sell_type_reverse_lookup(line_item),
        'costType': line_item.cost_type,
        'endDateTime': datetime_to_dfp_date(line_item.end),
        'targeting': transform_targeting_data_to_dfp(line_item.targeting),
    }

    if preserve_id:
        assert use_start
        if not line_item.id:
            logging.warning((
                "This line item (%s) does not have an existing id to "
                "query DFP with"
            ), line_item.id)
        elif not line_item.campaign_id:
            logging.warning((
                "This line item (%s) does not have an existing order "
                "id to query DFP with"
            ), line_item.id)
        else:
            # Set the existing DFP ids associating this line item object
            # with the line item and order already booked in DFP
            dfp_doc['id'] = line_item.id
            dfp_doc['orderId'] = line_item.campaign_id

    # indicate whether to check for immediate availability or not
    if use_start:
        assert line_item.start

        dfp_doc['startDateTimeType'] = 'USE_START_DATE_TIME'
        dfp_doc['startDateTime'] = datetime_to_dfp_date(line_item.start)
    else:
        dfp_doc['startDateTimeType'] = 'IMMEDIATELY'

    if line_item.creative_placeholder:
        dfp_doc['creativePlaceholders'] = line_item.creative_placeholder

    if line_item.primary_goal:
        dfp_doc['primaryGoal'] = {
            'goalType': line_item.primary_goal.goal_type,
            'unitType': line_item.primary_goal.unit_type,
            'units': line_item.primary_goal.units,
        }
    else:
        dfp_doc['primaryGoal'] = {
            'units': 100,
            'unitType': 'IMPRESSIONS'
        }

    if line_item.target_platform:
        dfp_doc['targetPlatform'] = line_item.target_platform

    return dfp_doc


def transform_creative_from_dfp(creative):
    """
    Convert dictionary-representation for a SUDS creative object into a
    Parselmouth representation of a Creative
    @param order: dict, dictionary-representation of a DFP SUDS response
        for a single creative
    @return: parselmouth.delivery.Creative
    """

    _, _, last_modified_datetime = _get_dfp_dates_from_object(creative)

    return Creative(
        advertiser_id=creative['advertiserId'],
        id=creative['id'],
        last_modified=last_modified_datetime,
        name=creative['name'],
        preview_url=creative['previewUrl'],
        size=creative['size']
    )


def transform_creative_to_dfp(creative):
    """
    Transform a Creative delivery object into
    a dictionary that will be accepted by dfp,

    @param creative: Creative
    @return: dict
    """
    return {
        'id': creative.id,
        'advertiserId': creative.advertiser_id,
        'name': creative.name,
        'previewUrl': creative.preview_url,
        'size': creative.size,
        'lastModifiedDateTime': datetime_to_dfp_date(creative.last_modified),
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth DFP Interface
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Imports
import logging
from datetime import timedelta
from pytz import timezone
from urllib import quote

# Parselmouth Imports
from parselmouth.adapters.abstract_interface import AbstractInterface
from parselmouth.delivery import Campaign
from parselmouth.delivery import LineItem
from parselmouth.exceptions import ParselmouthException
from parselmouth.targeting import AdUnit
from parselmouth.targeting import Custom
from parselmouth.targeting import Geography
from parselmouth.utils.dateutils import align_to_day

# Parselmouth Imports - Local DFP Adapter Imports
from parselmouth.adapters.dfp.constants import DFP_API_VERSION
from parselmouth.adapters.dfp.constants import DFP_QUERY_DEFAULTS
from parselmouth.adapters.dfp.constants import DFP_REPORT_METRIC_MAP
from parselmouth.adapters.dfp.client  import DFPClient
from parselmouth.adapters.dfp.utils import recursive_asdict
from parselmouth.adapters.dfp.delivery_utils import transform_line_item_from_dfp
from parselmouth.adapters.dfp.delivery_utils import transform_line_item_to_dfp
from parselmouth.adapters.dfp.delivery_utils import transform_forecast_line_item_to_dfp
from parselmouth.adapters.dfp.delivery_utils import transform_campaign_from_dfp
from parselmouth.adapters.dfp.delivery_utils import transform_creative_from_dfp


class DFPInterface(AbstractInterface):
    """
    Implementation of the DFP Ad service interface

    Serializes native Parselmouth objects into the PQL (Publisher
    Query Language) representations for querying objects
    """

    def __init__(self,
                 client_id,
                 client_secret,
                 refresh_token,
                 application_name,
                 network_code):
        """
        Constructor

        @param provider_config: child(parselmouth.config.ParselmouthConfig)
        """
        self.dfp_client = DFPClient(
            client_id,
            client_secret,
            refresh_token,
            application_name,
            network_code,
            version=DFP_API_VERSION,
        )

    def _convert_response_to_dict(self, dfp_data):
        """
        @param dfp_data: list(SUDS)
        @return: list(dict)
        """
        return [recursive_asdict(d) for d in dfp_data]

    def get_network_timezone(self):
        """
        Get the DFP network timezone for
        the host this client is connected to

        @return: pytz.timezone
        """
        dfp_data = self.dfp_client.get_network_data()
        data = recursive_asdict(dfp_data)
        return timezone(data['timeZone'])

    def get_campaign(self, campaign_id):
        """
        Return a campaign object given an id

        @param campaign_id: str, id of the campaign to return
        @return: parselmouth.delivery.Campaign
        """
        # Fetch the SUDS object and convert to a proper dictionary
        dfp_order = self.dfp_client.get_order(campaign_id)
        results = self._convert_response_to_dict(dfp_order)

        if len(results) == 0:
            raise ParselmouthException(
                "No results for campaign with id: {0}".format(campaign_id)
            )
        elif len(results) > 1:
            raise ParselmouthException(
                "More than one result for campaign with id: {0}".format(
                    campaign_id
                )
            )

        return transform_campaign_from_dfp(results[0])

    def get_campaigns(self,
                      order=DFP_QUERY_DEFAULTS['order'],
                      limit=None,
                      offset=DFP_QUERY_DEFAULTS['offset'],
                      **filter_kwargs):
        """
        Get campaigns on optional filters

        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: L{parselmouth.delivery.Campaign}

        Example:
            * Get an order by id: `get_campaigns(id=ORDER_ID)`
            * Get orders by advertiser id:
                `get_campaigns(advertiserId=ADVERTISER_ID, limit=10)`
        """
        # Fetch the SUDS object and convert to a proper dictionary
        dfp_orders = self.dfp_client.get_orders(
            order=order,
            limit=limit,
            offset=offset,
            **filter_kwargs
        )
        results = self._convert_response_to_dict(dfp_orders)

        campaigns = []
        for order in results:
            campaigns.append(
                transform_campaign_from_dfp(order)
            )
        return campaigns

    def get_line_item(self, line_item_id):
        """
        Return a line item object given an id

        @param line_item_id: str, id of the LineItem to return
        @return: parselmouth.delivery.LineItem
        """
        dfp_line_item = self.dfp_client.get_line_item(line_item_id)
        results = self._convert_response_to_dict(dfp_line_item)

        if len(results) == 0:
            raise ParselmouthException(
                "No results for line item with id: {0}".format(line_item_id)
            )
        elif len(results) > 1:
            raise ParselmouthException(
                "More than one result for line item with id: {0}".format(
                    line_item_id
                )
            )

        return transform_line_item_from_dfp(results[0])

    def get_line_items(self,
                       order=DFP_QUERY_DEFAULTS['order'],
                       limit=None,
                       offset=DFP_QUERY_DEFAULTS['offset'],
                       **filter_kwargs):
        """
        Get line items on optional filters

        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: L{parselmouth.delivery.LineItem}
        """

        # Fetch the SUDS object and convert to a proper dictionary
        dfp_line_items = self.dfp_client.get_line_items(
            order=order,
            limit=limit,
            offset=offset,
            **filter_kwargs
        )
        results = self._convert_response_to_dict(dfp_line_items)

        line_items = []
        for line_item in results:
            line_items.append(
                transform_line_item_from_dfp(line_item)
            )
        return line_items

    def get_campaign_line_items(self, campaign):
        """
        Get line items on optional filters

        @param campaign: Campaign|str
        @return: L{parselmouth.delivery.LineItem}
        """
        if isinstance(campaign, Campaign):
            _id = campaign.id
        else:
            _id = int(campaign)

        return self.get_line_items(orderId=_id)

    def get_line_item_available_inventory(self,
                                          line_item,
                                          use_start=False,
                                          preserve_id=False):
        """
        Get number of impressions available for line item

        @param line_item: LineItem
        @param use_start: bool, if False, checks availability from right
            now
        @param preserve_id: bool, communicate the id of the line item
            being forecasted to DFP. This is used in the case where we
            want to get a forecast on a line item in flight. This way, a
            domain doesn't need to have enough spare inventory to
            accommodate the two line items simultaneously. NOTE: If this
            is true then use_start is necessarily true.
        @return: int|None, number of available impressions
        """
        dfp_line_item = transform_forecast_line_item_to_dfp(
            line_item, use_start, preserve_id
        )
        logging.info(
            "Obtaining forecast data for line item %s",
            line_item.id or line_item
        )
        try:
            dfp_forecast = self.dfp_client.forecast_line_item(dfp_line_item)
        except Exception as e:
            raise ParselmouthException(e)

        forecast = recursive_asdict(dfp_forecast)
        if forecast and forecast.get('availableUnits'):
            available_units = int(forecast['availableUnits'])
        else:
            available_units = None

        logging.info(
            "%s available impressions",
            str(available_units)
        )
        return available_units

    def get_advertisers(self):
        """
        Queries dfp for all advertisers within their account

        @return: list(dict), returns a list of company dictionaries
        """
        brands = self.dfp_client.get_advertisers()

        if len(brands) == 0:
            raise ParselmouthException(
                "No brands found"
            )
        else:
            return brands

    def get_creative(self, creative_id):
        """
        Return a creative object given an id

        @param creative_id: str, id of the campaign to return
        @return: parselmouth.delivery.Creative
        """
        dfp_creative = self.dfp_client.get_creative(creative_id)
        creatives = self._convert_response_to_dict(dfp_creative)

        if len(creatives) == 0:
            raise ParselmouthException(
                "No results for creative with id: {0}".format(creative_id)
            )
        elif len(creatives) > 1:
            raise ParselmouthException(
                "More than one result for creative with id: {0}".format(
                    creative_id
                )
            )

        return transform_creative_from_dfp(creatives[0])

    def get_creatives(self,
                      order=DFP_QUERY_DEFAULTS['order'],
                      limit=None,
                      offset=DFP_QUERY_DEFAULTS['offset'],
                      **filter_kwargs):
        """
        Get creatives on optional filters

        @param order: str, PQL key to sort on (default=ID)
        @param limit: int, number of PQL results to return
        @param offset: int, page in a stream of PQL results to return
        @param filter_kwargs: dict, keyword arguments on which to filter
            PQL results
        @return: L{parselmouth.delivery.LineItem}
        """

        # Fetch the SUDS object and convert to a proper dictionary
        dfp_creatives = self.dfp_client.get_creatives(
            order=order,
            limit=limit,
            offset=offset,
            **filter_kwargs
        )
        results = self._convert_response_to_dict(dfp_creatives)

        creatives = []
        for creative in results:
            creatives.append(
                transform_creative_from_dfp(creative)
            )
        return creatives

    def get_line_item_creatives(self, line_item):
        """
        Return the creatives associated with a given line item

        @param line_item: str|int|parselmouth.delivery.LineItem,
            either the id of the lineitem or an object with the id
        @return: list(parselmouth.delivery.Creative)
        """

        if isinstance(line_item, LineItem):
            _id = line_item.id
        else:
            _id = int(line_item)
        creatives_to_fetch = self.dfp_client.get_line_item_creatives(_id)
        if len(creatives_to_fetch) == 0:
            raise ParselmouthException(
                "No creatives associated with line item: {0}".format(_id)
            )
        creatives = self.get_creatives(
            id=creatives_to_fetch,
            limit=len(creatives_to_fetch)
        )
        return creatives

    def get_geography_targets(self):
        """
        Get a list of DFP geography ids

        @return: list(dict)
        """
        geo_list = self.dfp_client.get_geography_targets()

        # The type field is in all caps, we want to
        # lower case this to be consistent with other systems
        output_list = []
        for item in geo_list:
            # Convert to Geography Target objects
            geo = Geography(
                id=item['id'],
                parent_id=item['countrycode'],
                type=item['type'].lower(),
                name=item['name'],
                external_name=quote(item['name'], ''),
            )
            output_list.append(geo)

        return output_list

    def get_adunit_targets(self):
        """
        Get a list of DFP adunit ids

        @return: list(dict)
        """
        adunit_list = self.dfp_client.get_adunit_targets()

        output_list = []
        for item in adunit_list:
            # Convert to Geography Target objects
            adunit = AdUnit(
                id=item['id'],
                parent_id=item['parentid'],
                name=item['name'],
                external_name=item['id'],
            )
            output_list.append(adunit)

        return output_list

    def _parse_custom_targets(self, custom_targets):
        """
        Helper function for converting native DFP custom targets into
        Parselmouth types

        @param custom_targets: list(dict)
        @return: list(parselmouth.targeting.Custom)
        """

        custom_targets = self._convert_response_to_dict(custom_targets)

        output_list = []
        for item in custom_targets:
            _external_id = item.get('name')
            if isinstance(_external_id, str):
                # Must be upper case to be consistent with redshift data
                _external_name = _external_id.upper()
            else:
                _external_name = _external_id

            display_name = item.get('displayName')
            if display_name == 'None':
                display_name = None

            output_list.append(
                Custom(
                    id=item['id'],
                    parent_id=item.get('customTargetingKeyId'),
                    name=item['name'],
                    type=item.get('type'),
                    external_id=_external_id,
                    external_name=_external_name,
                    descriptive_name=display_name,
                    # Fields needed for reconstructing dfp custom criterion
                    id_key='valueIds',
                    node_key='CustomCriteria',
                )
            )

        return output_list

    def get_custom_targets(self, key_name=None, value_name=None):
        """
        Get all key and value data from the custom
        targeting service

        @param key_name: str|None, if present only return
            targeting with the given key name
        @param value_name: str|None, if present only return
            targeting with the given value name
        @return: list(parselmouth.targeting.Custom)
        """
        dfp_custom_targets = self.dfp_client.get_custom_targets(
            key_name, value_name,
        )
        return self._parse_custom_targets(dfp_custom_targets)

    def create_custom_target(self, key, value):
        """
        Add a custom target to this domain's ad service provider

        @param key: parselmouth.targeting.Custom
        @param value: parselmouth.targeting.Custom
        @return: list(parselmouth.targeting.Custom)
        """
        assert isinstance(key, Custom)
        assert isinstance(value, Custom)

        dfp_custom_targets = self.dfp_client.create_custom_target(
            key_name=key.name,
            key_type=key.type,
            key_display_name=key.descriptive_name,
            value_name=value.name,
            value_match_type=value.type,
            value_display_name=value.descriptive_name,
        )
        return self._parse_custom_targets(dfp_custom_targets)

    def get_line_item_report(self,
                             start,
                             end,
                             columns=['ad_impressions']):
        """
        Get the number of impressions served by DFP
        for all line items between the two datetimes.

        Note: These datetimes are assumed to be in the timezone
        of the host and will be aligned to day

        Note: This API has a strange behavior where the date
        inputs only allow for a calendar day.
        This means start=1/1, end=1/1 will query against
        the entire day of 1/1

        @param start: datetime,
        @param end: datetime,
        @param columns: list(str), list of columns to include
        @return: list(dict)
        """
        assert [DFP_REPORT_METRIC_MAP[c] for c in columns]

        dfp_start = align_to_day(start)
        # You must subtract one day because the entire end
        # date will be included in the query result
        dfp_end = align_to_day(end) - timedelta(days=1)

        results = self.dfp_client.generate_report(
            dimensions=[
                DFP_REPORT_METRIC_MAP['line_item_id'],
                DFP_REPORT_METRIC_MAP['line_item_name'],
            ],
            columns=[DFP_REPORT_METRIC_MAP[c] for c in columns],
            date_range_type='CUSTOM_DATE',
            start_date=dfp_start,
            end_date=dfp_end,
        )

        result = []
        for _item in results:
            # Translate to chartbeat keys
            doc = {}
            for _key, dfp_key in DFP_REPORT_METRIC_MAP.items():
                if _item.get(dfp_key) is not None and _key in columns:
                    try:
                        doc[_key] = int(_item[dfp_key])
                    except ValueError:
                        doc[_key] = _item[dfp_key]
                elif _item.get(dfp_key) is not None:
                    doc[_key] = _item[dfp_key]

            result.append(doc)

        return result

    def update_line_items(self, line_items):
        """
        Update multiple Line Item for a provider

        @param line_items: L{parselmouth.delivery.LineItem}
        """
        # Convert line items into native format for the dfp client to use
        native_line_items = [
            transform_line_item_to_dfp(line_item)
            for line_item in line_items
        ]
        self.dfp_client.update_line_items(native_line_items)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Parselmouth - Delivery Models
Base Interfaces for delivery items
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
from datetime import datetime

# Parselmouth Imports
from parselmouth.targeting import TargetingData
from parselmouth.model import ObjectModel


class Cost(ObjectModel):
    """
    Container class for modeling Campaign and Line Item Budgets and
    Costs
    """

    def __init__(self,
                 budget_micro_amount,
                 budget_currency_code):
        """
        @param budget_micro_amount: str|int
        @param budget_currency_code: str
        """
        self.budget_micro_amount = float(budget_micro_amount)
        self.budget_currency_code = budget_currency_code


class Goal(ObjectModel):
    """
    Container class for modeling delivery goals
    """

    def __init__(self,
                 goal_type=None,
                 unit_type=None,
                 units=None):

        self.goal_type = goal_type
        self.unit_type = unit_type
        self.units = units


class Stats(ObjectModel):
    """
    Class for tracking a container's delivery performance
    """

    def __init__(self,
                 impressions,
                 clicks,
                 video_starts=0,
                 video_completions=0,
                 **kwargs):
        """
        @param impressions: str|int, number of impressions delivered
        @param clicks: str|int, number of clicks delivered
        """
        self.impressions = int(impressions)
        self.clicks = int(clicks)
        self.video_starts = int(video_starts)
        self.video_completions = int(video_completions)
        if self.clicks and self.impressions:
            self.click_through_rate = self.clicks / self.impressions
        else:
            self.click_through_rate = 0


class DeliveryMeta(ObjectModel):
    """
    Class for tracking delivery expectations
    """

    def __init__(self,
                 stats,
                 delivery_rate_type,
                 actual_delivery_percent,
                 expected_delivery_percent,
                 **kwargs):
        """
        @param stats: parselmouth.delivery.Stats
        @param delivery_rate_type: str
        @param actual_delivery_percent: str|float
        @param expected_delivery_percent: str|float
        """

        assert isinstance(stats, Stats)

        self.stats = stats
        self.delivery_rate_type = delivery_rate_type
        self.actual_delivery_percent = float(actual_delivery_percent)
        self.expected_delivery_percent = float(expected_delivery_percent)
        if actual_delivery_percent > 0 and expected_delivery_percent > 0:
            self.pace = \
                self.actual_delivery_percent / self.expected_delivery_percent
        else:
            self.pace = 0


class Creative(ObjectModel):
    """
    Abstract Class for creative containers
    A creative is a specific advertisement, such as an image file, video
    file, or other content that gets delivered to end users. A creative
    can be associated with one or more line items.
    """

    def __init__(self,
                 advertiser_id=None,
                 type=None,
                 id=None,
                 last_modified=None,
                 name=None,
                 preview_url=None,
                 size=None):

        assert isinstance(last_modified, datetime) or last_modified is None

        self.advertiser_id = advertiser_id
        self.type = type
        self.id = id
        self.last_modified = last_modified
        self.name = name
        self.preview_url = preview_url
        self.size = size


class LineItem(ObjectModel):
    """
    Abstract Class for Line Items
    A line item represents an atomically purchased set of ads over a
    period of time. An individual line item has its own targeting,
    impression goal, and cost. This is most basic unit that is purchased
    when building an ad campaign for a particular campaign. Every line
    item is contained within a campaign or order. A line item contains
    one or more creatives.
    """

    def __init__(self,
                 budget=None,
                 cost_per_unit=None,
                 cost_type=None,
                 delivery=None,
                 domain=None,
                 end=None,
                 id=None,
                 last_modified=None,
                 last_modified_by=None,
                 type=None,
                 name=None,
                 campaign_id=None,
                 campaign_name=None,
                 primary_goal=None,
                 start=None,
                 status=None,
                 targeting=None,
                 value_cost_per_unit=None,
                 creative_placeholder=None,
                 target_platform=None):

        assert isinstance(start, datetime) or start is None
        assert isinstance(last_modified, datetime) or last_modified is None
        assert isinstance(end, datetime) or end is None
        assert isinstance(budget, Cost) or budget is None
        assert isinstance(cost_per_unit, Cost) or cost_per_unit is None
        assert isinstance(value_cost_per_unit, Cost) or value_cost_per_unit is None
        assert isinstance(primary_goal, Goal) or primary_goal is None
        assert isinstance(delivery, DeliveryMeta) or delivery is None
        assert isinstance(targeting, TargetingData) or targeting is None

        if start and end:
            assert start <= end

        self.budget = budget
        self.cost_per_unit = cost_per_unit
        self.cost_type = cost_type
        self.delivery = delivery
        self.domain = domain
        self.end = end
        self.id = id
        self.last_modified = last_modified
        self.last_modified_by = last_modified_by
        self.type = type
        self.name = name
        self.campaign_id = campaign_id
        self.campaign_name = campaign_name
        self.primary_goal = primary_goal
        self.start = start
        self.status = status
        self.targeting = targeting
        self.value_cost_per_unit = value_cost_per_unit
        self.creative_placeholder = creative_placeholder
        self.target_platform = target_platform

    @property
    def stats(self):
        """
        Proxy to the self.delivery.stats object that reports a line
        item's delivered impressions and clicks. This is mainly here so
        we can treat Campaign.stats and LineItem.stats the same...
        Makes self.delivery.stats accessible by just self.stats
        """
        return self.delivery.stats


class Campaign(ObjectModel):
    """
    Abstract Class for Campaigns
    A campaign is an object with a start and end date, and is a
    container for one or more line items.
    """

    def __init__(self,
                 advertiser_id=None,
                 agency_id=None,
                 creator_id=None,
                 currency_code=None,
                 domain=None,
                 end=None,
                 external_campaign_id=None,
                 id=None,
                 last_modified=None,
                 last_modified_by=None,
                 name=None,
                 start=None,
                 status=None,
                 stats=None,
                 total_budget=None,
                 line_items=None):

        assert isinstance(start, datetime) or start is None
        assert isinstance(last_modified, datetime) or last_modified is None
        assert isinstance(end, datetime) or end is None
        assert isinstance(stats, Stats) or stats is None

        if start and end:
            assert start <= end

        self.advertiser_id = advertiser_id
        self.agency_id = agency_id
        self.creator_id = creator_id
        self.currency_code = currency_code
        self.domain = domain
        self.end = end
        self.external_campaign_id = external_campaign_id
        self.id = id
        self.last_modified = last_modified
        self.last_modified_by = last_modified_by
        self.name = name
        self.start = start
        self.status = status
        self.stats = stats
        self.total_budget = total_budget
        self.line_items = line_items



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Parselmouth utilities - Date Utilites
"""

# Future-proof
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Library Imports
import time
from datetime import timedelta


def align_to_day(d, fix_utc_offset=False):
    """Align a datetime to a day boundary.

    @param d: datetime to align
    @param fix_utc_offset: bool, converts `d` from America/New_York to UTC
    @return datetime
    """
    if not d:
        return d

    hour_offset = 0
    if fix_utc_offset:
        hour_offset = get_et_utc_offset()

    return d - timedelta(hours=d.hour + hour_offset,
                         minutes=d.minute,
                         seconds=d.second,
                         microseconds=d.microsecond)


def get_et_utc_offset():
    """
    Returns number of hours we need to substract from a timestamp
    to account for utc/et issues in our code.
    """
    if is_localtime_dst():
        return 4
    return 5


def is_localtime_dst():
    """
    Check if the local time is using daylight savings.

    @return: bool
    """
    current_time = time.localtime()
    return True if current_time.tm_isdst else False
